import numpy as np
np.set_printoptions(linewidth=320)
from numpy import zeros, ones, mod, conj, array, c_, r_, linalg, Inf, complex128
from itertools import product
from numpy.linalg import solve, inv
from scipy.sparse.linalg import factorized
from scipy.sparse import issparse, csc_matrix as sparse

np.set_printoptions(linewidth=1000000, )

# Set the complex precision to use
complex_type = complex128


def zpf(Vbus, Sbus, Ibus, Ybus, pq, pv, ref, pqpv, tol=1e-9, max_ter=100):
    """

    Args:
        Vbus:
        Sbus:
        Ibus:
        Ybus:
        pq:
        pv:
        ref:
        pqpv:
        tol:

    Returns:

    """

    # reduced impedance matrix
    Zred = factorized(Ybus[pqpv, :][:, pqpv])

    # slack currents
    Ivd = Ybus[pqpv, :][:, ref].dot(Vbus[ref])
    print('Ivd', np.vstack(Ivd))

    # slack voltages influence
    Ck = Zred(Ivd)
    print('Ck', np.vstack(Ck))
    # make a copy of the voltage for convergence control
    Vprev = Vbus[pqpv].copy()

    # Voltage module in the pv nodes
    Vpv = abs(Vbus[pv])

    # admittance matrix to compute the reactive power
    Ybus_pv = Ybus[pv, :][:, pv]

    # approximate the currents with the current voltage solution
    Ik = conj(Sbus[pqpv] / Vprev) + Ibus[pqpv]
    print('Sred', np.vstack(Sbus[pqpv]))
    print('Ik', np.vstack(Ik))

    # compute the new voltage solution
    Vk = Zred(Ik) - Ck
    print('Vk', np.vstack(Vk))

    # compute the voltage solution maximum difference
    diff = max(abs(Vprev - Vk))

    iter = 1
    while diff > tol and iter < max_ter:

        # make a copy of the voltage for convergence control
        Vprev = Vk

        # approximate the currents with the current voltage solution
        Ik = conj(Sbus[pqpv] / Vprev) + Ibus[pqpv]

        # compute the new voltage solution
        Vk = Zred(Ik) - Ck
        print(iter, 'Vk', Vk)
        print()
        # tune PV nodes
        #  ****** USE A reduced pv, pv, pqpv mapping!
        # Vk[pv] *= Vpv / abs(Vk[pv])
        # Qpv = (Vk * conj(Ybus[pv, :][:, pv].dot(Vk) - Ibus))[pv].imag
        # Sbus[pv] = Sbus[pv].real + 1j * Qpv

        # compute the voltage solution maximum difference
        diff = max(abs(Vprev - Vk))

        # Assign the reduced voltage solution to the complete voltage solution
        # voltage = Vbus.copy()  # the slack voltages are kept
        # voltage[pqpv] = Vk
        # compute the power mismatch: this is the true equation solution check
        # Scalc = voltage * conj(Ybus * voltage - Ibus)
        # mis = Scalc - Sbus  # complex power mismatch
        # diff = linalg.norm(r_[mis[pv].real, mis[pq].real, mis[pq].imag], Inf)

        iter += 1

    # Assign the reduced voltage solution to the complete voltage solution
    voltage = Vbus.copy()  # the slack voltages are kept
    voltage[pqpv] = Vk

    print(iter, 'voltage:\n', np.vstack(voltage))
    print()

    # compute the power mismatch: this is the true equation solution check
    Scalc = voltage * conj(Ybus * voltage - Ibus)
    mis = Scalc - Sbus  # complex power mismatch
    normF = linalg.norm(r_[mis[pv].real, mis[pq].real, mis[pq].imag], Inf)

    print('Iter: ', iter)

    return voltage, normF


if __name__ == '__main__':
    from GridCal.Engine.calculation_engine import *

    grid = MultiCircuit()
    grid.load_file('lynn5buspq.xlsx')
    # grid.load_file('IEEE30.xlsx')
    # grid.load_file('C:\\Users\\spenate\Documents\\PROYECTOS\\Monash\\phase0\\Grid\\Monash University.xlsx')
    # grid.load_file('D:\\GitHub\\GridCal\\Grids_and_profiles\\grids\\IEEE_14.xlsx')
    grid.compile()

    circuit = grid.circuits[0]

    print('\nYbus:\n', circuit.power_flow_input.Ybus.todense())
    print('\nYseries:\n', circuit.power_flow_input.Yseries.todense())
    print('\nYshunt:\n', circuit.power_flow_input.Yshunt)
    print('\nSbus:\n', circuit.power_flow_input.Sbus)
    print('\nIbus:\n', circuit.power_flow_input.Ibus)
    print('\nVbus:\n', circuit.power_flow_input.Vbus)
    print('\ntypes:\n', circuit.power_flow_input.types)
    print('\npq:\n', circuit.power_flow_input.pq)
    print('\npv:\n', circuit.power_flow_input.pv)
    print('\nvd:\n', circuit.power_flow_input.ref)

    import time

    print('Z-Gaus-Seidel')
    start_time = time.time()

    v, err = zpf(Vbus=circuit.power_flow_input.Vbus,
                 Sbus=circuit.power_flow_input.Sbus,
                 Ibus=circuit.power_flow_input.Ibus,
                 Ybus=circuit.power_flow_input.Ybus,
                 pq=circuit.power_flow_input.pq,
                 pv=circuit.power_flow_input.pv,
                 ref=circuit.power_flow_input.ref,
                 pqpv=circuit.power_flow_input.pqpv)

    print("--- %s seconds ---" % (time.time() - start_time))

    print('V module:\t', np.abs(v))
    print('V angle: \t', np.angle(v))
    print('error: \t', err)

    # check the HELM solution: v against the NR power flow
    print('\nNR')
    options = PowerFlowOptions(SolverType.NR, verbose=False, robust=False, tolerance=1e-9)
    power_flow = PowerFlow(grid, options)

    start_time = time.time()
    power_flow.run()
    print("--- %s seconds ---" % (time.time() - start_time))
    vnr = circuit.power_flow_results.voltage

    print('V module:\t', np.abs(vnr))
    print('V angle: \t', np.angle(vnr))
    print('error: \t', circuit.power_flow_results.error)

    # check
    print('\ndiff:\t', v - vnr)


"""
Method implemented from the article:
Online voltage stability assessment for load areas based on the holomorphic embedding method
by Chengxi Liu, Bin Wang, Fengkai Hu, Kai Sun and Claus Leth Bak

Implemented by Santiago Peñate Vera 2018
"""
import numpy as np
np.set_printoptions(linewidth=32000, suppress=False)
from numpy import zeros, ones, mod, angle, conj, array, c_, r_, linalg, Inf, complex128
from numpy.linalg import solve
from scipy.sparse.linalg import factorized
from scipy.sparse import lil_matrix
from scipy.sparse import hstack as hstack_s, vstack as vstack_s

# Set the complex precision to use
complex_type = complex128


def prepare_system_matrices(Ybus, Vbus, bus_idx, pqpv, pq, pv, ref):
    """
    Prepare the system matrices
    :param Ybus: Admittanche matrix
    :param Vbus: Node complex voltage vector (initial set voltages)
    :param pqpv: list of pq and pv bus indices
    :param ref: list of slack node indices
    :return: System matrix, initial voltage seed, initial inverse voltage seed
    """
    n_bus = len(Vbus)
    n_bus2 = 2 * n_bus
    npv = len(pv)
    # ##################################################################################################################
    # Compute the starting voltages
    # ##################################################################################################################

    # System matrix
    A = lil_matrix((n_bus2, n_bus2))  # lil matrices are faster to populate

    # Expanded slack voltages
    Vslack = zeros(n_bus2)

    # Populate A
    for a in pqpv:  # rows
        for ii in range(Ybus.indptr[a], Ybus.indptr[a + 1]):  # columns in sparse format
            b = Ybus.indices[ii]

            A[2 * a + 0, 2 * b + 0] = Ybus[a, b].real
            A[2 * a + 0, 2 * b + 1] = -Ybus[a, b].imag
            A[2 * a + 1, 2 * b + 0] = Ybus[a, b].imag
            A[2 * a + 1, 2 * b + 1] = Ybus[a, b].real

    # set vd elements
    for a in ref:
        A[a * 2, a * 2] = 1.0
        A[a * 2 + 1, a * 2 + 1] = 1.0

        Vslack[a * 2] = Vbus[a].real
        Vslack[a * 2 + 1] = Vbus[a].imag

    # Solve starting point voltages
    Vst_expanded = factorized(A.tocsc())(Vslack)

    # Invert the voltages obtained: Get the complex voltage and voltage inverse vectors
    Vst = Vst_expanded[2 * bus_idx] + 1j * Vst_expanded[2 * bus_idx + 1]
    Wst = 1.0 / Vst

    # ##################################################################################################################
    # Compute the final system matrix
    # ##################################################################################################################

    # System matrices
    B = lil_matrix((n_bus2, 3 * npv))
    C = lil_matrix((3 * npv, n_bus2))
    D = lil_matrix((3 * npv, 3 * npv))

    for i, a in enumerate(pv):
        # "a" is the actual bus index
        # "i" is the number of the pv bus in the pv buses list

        B[2 * a + 0, 3 * i + 2] = Wst[a].imag
        B[2 * a + 1, 3 * i + 2] = Wst[a].real

        C[3 * i + 0, 2 * a + 0] = Wst[a].real
        C[3 * i + 0, 2 * a + 1] = -Wst[a].imag
        C[3 * i + 1, 2 * a + 0] = Wst[a].real
        C[3 * i + 1, 2 * a + 1] = Wst[a].imag
        C[3 * i + 2, 2 * a + 0] = Vst[a].real
        C[3 * i + 2, 2 * a + 1] = Vst[a].imag

        D[3 * i + 0, 3 * i + 0] = Vst[a].real
        D[3 * i + 0, 3 * i + 1] = -Vst[a].imag
        D[3 * i + 1, 3 * i + 0] = Vst[a].imag
        D[3 * i + 1, 3 * i + 1] = Vst[a].real

    Asys = vstack_s([
                    hstack_s([A, B]),
                    hstack_s([C, D])
                    ], format="csc")

    return Asys, Vst, Wst


def calc_W(n, V, W, pq):
    """
    Calculation of the inverse coefficients W.
    @param n: Order of the coefficients
    @param V: Structure of voltage coefficients (Ncoeff x nbus elements)
    @param W: Structure of inverse voltage coefficients (Ncoeff x nbus elements)
    @return: Array of inverse voltage coefficients for the order n
    """

    if n == 0:
        res = 1.0 / conj(V[0, :][pq])
    else:
        l = array(range(n))
        res = -(W[l, :][:, pq] * V[n - l, :][:, pq]).sum(axis=0)
        res /= conj(V[0, :][pq])

    return res


def get_rhs(n, V, W, Q, Vbus, Vst, Sbus, Pbus, nsys, nbus2, pv, pq, pvpos):
    """
    Right hand side
    :param n: order of the coefficients
    :param V: Voltage coefficients (order, all buses)
    :param W: Inverse voltage coefficients (order, pv buses)
    :param Q: Reactive power coefficients  (order, pv buses)
    :param Vbus: Initial bus estimate (only used to pick the PV buses set voltage)
    :param Vst: Start voltage due to slack injections
    :param Pbus: Active power injections (all the buses)
    :param nsys: number of rows or cols in the system matrix A
    :param nbus2: two times the number of buses
    :param pv: list of pv indices in the grid
    :param pvpos: array from 0..npv
    :return: right hand side vector to solve the coefficients of order n
    """
    rhs = zeros(nsys)
    m = array(range(1, n), dtype=int)
    # ##################################################################################################################
    # PQ nodes
    # ##################################################################################################################

    f1 = conj(Sbus[pq] * W[:, pq][n - 1, :])
    idx1 = 2 * pq
    rhs[idx1 + 0] = f1.real
    rhs[idx1 + 1] = f1.imag

    # ##################################################################################################################
    # PV nodes
    # ##################################################################################################################
    # Compute convolutions
    QW_convolution = (Q[n - m, :] * W[m, :][:, pv].conjugate()).sum(axis=0)  # only pv nodes
    WV_convolution = (W[n - m, :] * V[m, :]).sum(axis=0)  # all nodes
    VV_convolution = (V[m, :][:, pv] * V[n - m, :][:, pv].conjugate()).sum(axis=0)  # only pv nodes

    # compute the formulas
    f2 = Pbus[pv] * W[:, pv][n-1, :] + QW_convolution

    epsilon = -0.5 * VV_convolution
    if n == 1:
        epsilon += 0.5 * (abs(Vbus[pv]) ** 2 - abs(Vst[pv]) ** 2)

    # Assign the values to the right hand side vector
    idx2 = 2 * pv
    idx3 = 3 * pvpos + nbus2

    rhs[idx2 + 0] = f2.real
    rhs[idx2 + 1] = f2.imag

    if len(idx3) > 0:

        rhs[idx3 + 0] = -WV_convolution.real[pv]
        rhs[idx3 + 1] = -WV_convolution.imag[pv]
        rhs[idx3 + 2] = epsilon.real

    else:

        # No PV nodes
        pass

    return rhs


def assign_solution(x, bus_idx, pvpos, pq, pv, nbus):
    """
    Assign the solution vector to the appropriate coefficients
    :param x: solution vector
    :param bus_idx: array from 0..nbus-1
    :param nbus2: two times the number of buses (integer)
    :param pvpos: array from 0..npv
    :return: Array of:
            - voltage coefficients
            - voltage inverse coefficients
            - reactive power
            of order n
    """

    nbus2 = 2 * nbus

    # declare a row of inverse voltage coefficients
    w = np.zeros(nbus, dtype=complex_type)

    # assign the voltage coefficients
    v = x[2 * bus_idx] + 1j * x[2 * bus_idx + 1]

    if len(pvpos) > 0:

        # assign the inverse voltage coefficients of the PV nodes
        w[pv] = x[nbus2 + 3 * pvpos] + 1j * x[nbus2 + 3 * pvpos + 1]

        # assign the reactive power coefficients of the PV nodes
        q = x[nbus2 + 3 * pvpos + 2]

    else:

        # No PV nodes

        w = zeros(nbus, dtype=complex_type)

        q = zeros(0)

    return v, w, q


def pade_approximation(n, an, s=1):
    """
    Computes the n/2 pade approximant of the series an at the approximation
    point s

    Arguments:
        an: coefficient matrix, (number of coefficients, number of series)
        n:  order of the series
        s: point of approximation

    Returns:
        pade approximation at s
    """
    nn = int(n / 2)
    if mod(nn, 2) == 0:
        nn -= 1

    L = nn
    M = nn

    an = np.ndarray.flatten(an)
    rhs = an[L + 1:L + M + 1]

    C = zeros((L, M), dtype=complex_type)
    for i in range(L):
        k = i + 1
        C[i, :] = an[L - M + k:L + k]

    try:
        b = solve(C, -rhs)  # bn to b1
    except:
        return 0, zeros(L + 1, dtype=complex_type), zeros(L + 1, dtype=complex_type)

    b = r_[1, b[::-1]]  # b0 = 1

    a = zeros(L + 1, dtype=complex_type)
    a[0] = an[0]
    for i in range(L):
        val = complex_type(0)
        k = i + 1
        for j in range(k + 1):
            val += an[k - j] * b[j]
        a[i + 1] = val

    p = complex_type(0)
    q = complex_type(0)
    for i in range(L + 1):
        p += a[i] * s ** i
        q += b[i] * s ** i

    return p / q, a, b


def helm_(Vbus, Sbus, Ibus, Ybus, pq, pv, ref, pqpv, tol=1e-9):
    """
    Helm Method
    :param Vbus: voltages array
    :param Sbus: Power injections array
    :param Ibus: Currents injection array
    :param Ybus: System admittance matrix
    :param pq: list of pq node indices
    :param pv: list of pv node indices
    :param ref: list of slack node indices
    :param pqpv: list of pq and pv node indices sorted
    :param tol: tolerance
    :return: Voltage array and the power mismatch
    """

    nbus = len(Vbus)
    npv = len(pv)
    bus_idx = array(range(nbus), dtype=int)
    pvpos = array(range(npv))

    # Prepare system matrices
    Asys, Vst, Wst = prepare_system_matrices(Ybus, Vbus, bus_idx, pqpv, pq, pv, ref)

    # Factorize the system matrix
    Afact = factorized(Asys)

    # get the shape
    nsys = Asys.shape[0]

    # declare the active power injections
    Pbus = Sbus.real

    # declare the matrix of coefficients: [order, bus index]
    V = zeros((1, nbus), dtype=complex_type)

    # Declare the inverse voltage coefficients: [order, pv bus index]
    W = zeros((1, nbus), dtype=complex_type)

    # Reactive power coefficients on the PV nodes: [order, pv bus index]
    Q = zeros((1, npv), dtype=double)

    # Assign the initial values
    V[0, :] = Vst
    W[0, :] = Wst
    Q[0, :] = zeros(npv)

    error = list()

    for n in range(1, 20):

        # Compute the free terms
        rhs = get_rhs(n=n, V=V, W=W, Q=Q,
                      Vbus=Vbus, Vst=Vst,
                      Sbus=Sbus,
                      Pbus=Pbus, nsys=nsys,
                      nbus2=2 * nbus,
                      pv=pv, pq=pq,
                      pvpos=pvpos)

        # Solve the linear system Asys x res = rhs
        res = Afact(rhs)

        # get the new rows of coefficients
        v, w, q = assign_solution(x=res, bus_idx=bus_idx, pvpos=pvpos, pq=pq, pv=pv, nbus=nbus)

        # Add coefficients row
        V = np.vstack((V, v))
        Q = np.vstack((Q, q))

        # compute the W coefficients for the PQ Nodes and update the whole W structure
        w[pq] = calc_W(n, V, W, pq)
        W = np.vstack((W, w))

        #     print('\nn:', n)
        #     print('RHS:\n', rhs)
        #     print('X:\n', res)
        #
        # print('V:\n', V)
        # print('W:\n', W)
        # print('Q:\n', Q)

        # Perform the Padè approximation
        # NOTE: Apparently the padé approximation is equivalent to the bare sum of coefficients !!
        # voltage = zeros(nbus, dtype=complex_type)
        # for i in range(nbus):
        #     voltage[i], _, _ = pade_approximation(n, V[:, i])

        voltage = V.sum(axis=0)

        # Calculate the error and check the convergence
        Scalc = voltage * conj(Ybus * voltage)
        mismatch = Scalc - Sbus  # complex power mismatch
        power_mismatch_ = r_[mismatch[pv].real, mismatch[pq].real, mismatch[pq].imag]

        # check for convergence
        normF = linalg.norm(power_mismatch_, Inf)

        if npv > 0:
            a = linalg.norm(mismatch[pv].real, Inf)
        else:
            a = 0
        b = linalg.norm(mismatch[pq].real, Inf)
        c = linalg.norm(mismatch[pq].imag, Inf)
        error.append([a, b, c])

    print('V:\n', V)
    print('W:\n', W)
    print('Q:\n', Q)

    err_df = pd.DataFrame(array(error), columns=['PV_real', 'PQ_real', 'PQ_imag'])
    err_df.plot(logy=True)

    return voltage, normF


def res_2_df(V, Sbus, tpe):
    """
    Create dataframe to display the results nicely
    :param V: Voltage complex vector
    :param Sbus: Power complex vector
    :param tpe: Types
    :return: Pandas DataFrame
    """
    vm = abs(V)
    va = angle(V)

    d = {1: 'PQ', 2: 'PV', 3: 'VD'}

    tpe_str = array([d[i] for i in tpe], dtype=object)
    data = c_[tpe_str, Sbus.real, Sbus.imag, vm, va]
    cols = ['Type', 'P', 'Q', '|V|', 'angle']
    df = pd.DataFrame(data=data, columns=cols)

    return df


if __name__ == '__main__':
    from GridCal.Engine.calculation_engine import *
    from matplotlib import pyplot as plt

    grid = MultiCircuit()
    # grid.load_file('lynn5buspq.xlsx')
    # grid.load_file('lynn5buspv.xlsx')
    # grid.load_file('IEEE30.xlsx')
    # grid.load_file('/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE 14.xlsx')
    grid.load_file('/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE39.xlsx')
    # grid.load_file('/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/1354 Pegase.xlsx')

    grid.compile()

    circuit = grid.circuits[0]

    print('\nYbus:\n', circuit.power_flow_input.Ybus.todense())
    print('\nSbus:\n', circuit.power_flow_input.Sbus)
    print('\nIbus:\n', circuit.power_flow_input.Ibus)
    print('\nVbus:\n', circuit.power_flow_input.Vbus)
    print('\ntypes:\n', circuit.power_flow_input.types)
    print('\npq:\n', circuit.power_flow_input.pq)
    print('\npv:\n', circuit.power_flow_input.pv)
    print('\nvd:\n', circuit.power_flow_input.ref)

    start_time = time.time()

    v, err = helm_(Vbus=circuit.power_flow_input.Vbus,
                   Sbus=circuit.power_flow_input.Sbus,
                   Ibus=circuit.power_flow_input.Ibus,
                   Ybus=circuit.power_flow_input.Ybus,
                   pq=circuit.power_flow_input.pq,
                   pv=circuit.power_flow_input.pv,
                   ref=circuit.power_flow_input.ref,
                   pqpv=circuit.power_flow_input.pqpv)

    print('HEM:')
    print("--- %s seconds ---" % (time.time() - start_time))
    print('Results:\n', res_2_df(v, circuit.power_flow_input.Sbus, circuit.power_flow_input.types))
    print('error: \t', err)

    # check the HELM solution: v against the NR power flow
    print('\nNR')
    options = PowerFlowOptions(SolverType.NR, verbose=False, robust=False, tolerance=1e-9, control_q=False)
    power_flow = PowerFlow(grid, options)

    start_time = time.time()
    power_flow.run()
    print("--- %s seconds ---" % (time.time() - start_time))
    vnr = circuit.power_flow_results.voltage

    print('Results:\n', res_2_df(vnr, circuit.power_flow_input.Sbus, circuit.power_flow_input.types))
    print('error: \t', circuit.power_flow_results.error)

    # check
    print('\ndiff:\t', v - vnr)

    plt.show()



# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.
import time
import multiprocessing
from PySide2.QtCore import QThread, Signal

from GridCal.Engine.basic_structures import Logger
from GridCal.Engine.Core.multi_circuit import MultiCircuit
from GridCal.Engine.Simulations.PowerFlow.power_flow_options import PowerFlowOptions
from GridCal.Engine.Simulations.LinearFactors.ptdf_analysis import get_ptdf_variations, power_flow_worker, PtdfGroupMode
from GridCal.Engine.Simulations.LinearFactors.ptdf_results import PTDFResults
from GridCal.Engine.Core.snapshot_pf_data import compile_snapshot_circuit

########################################################################################################################
# Optimal Power flow classes
########################################################################################################################


class PTDFOptions:

    def __init__(self, group_mode: PtdfGroupMode = PtdfGroupMode.ByGenLoad,
                 power_increment=100.0, use_multi_threading=False):
        """
        Power Transfer Distribution Factors' options
        :param group_mode: Grouping type
        :param power_increment: Amount of power to change in MVA
        :param use_multi_threading: use multi-threading?
        """
        self.group_mode = group_mode

        self.power_increment = power_increment

        self.use_multi_threading = use_multi_threading


class PTDF(QThread):
    progress_signal = Signal(float)
    progress_text = Signal(str)
    done_signal = Signal()
    name = 'PTDF'

    def __init__(self, grid: MultiCircuit, options: PTDFOptions, pf_options: PowerFlowOptions, opf_results=None):
        """
        Power Transfer Distribution Factors class constructor
        @param grid: MultiCircuit Object
        @param options: OPF options
        """
        QThread.__init__(self)

        # Grid to run
        self.grid = grid

        # Options to use
        self.options = options

        # power flow options
        self.pf_options = pf_options

        self.opf_results = opf_results

        # OPF results
        self.results = None

        # set cancel state
        self.__cancel__ = False

        self.all_solved = True

        self.elapsed = 0.0

        self.logger = Logger()

    def ptdf(self, circuit: MultiCircuit, options: PowerFlowOptions, group_mode: PtdfGroupMode, power_amount,
             text_func=None, prog_func=None):
        """
        Power Transfer Distribution Factors analysis
        :param circuit: MultiCircuit instance
        :param options: power flow options
        :param group_mode: group mode
        :param power_amount: amount o power to vary in MW
        :param text_func: text function to display progress
        :param prog_func: progress function to display progress [0~100]
        :return:
        """

        if text_func is not None:
            text_func('Compiling...')

        # compile to arrays
        numerical_circuit = compile_snapshot_circuit(circuit=circuit,
                                                     apply_temperature=options.apply_temperature_correction,
                                                     branch_tolerance_mode=options.branch_impedance_tolerance_mode,
                                                     opf_results=self.opf_results)

        calculation_inputs = numerical_circuit.split_into_islands(ignore_single_node_islands=options.ignore_single_node_islands)

        # compute the variations
        delta_of_power_variations = get_ptdf_variations(circuit=circuit,
                                                        numerical_circuit=numerical_circuit,
                                                        group_mode=group_mode,
                                                        power_amount=power_amount)

        # declare the PTDF results
        results = PTDFResults(n_variations=len(delta_of_power_variations) - 1,
                              n_br=numerical_circuit.nbr,
                              n_bus=numerical_circuit.nbus,
                              br_names=numerical_circuit.branch_names,
                              bus_names=numerical_circuit.bus_names,
                              bus_types=numerical_circuit.bus_types)

        if text_func is not None:
            text_func('Running PTDF...')

        nvar = len(delta_of_power_variations)
        for v, variation in enumerate(delta_of_power_variations):

            # this super strange way of calling a function is done to maintain the same
            # call format as the multi-threading function
            returns = dict()
            power_flow_worker(variation=0,
                              nbus=numerical_circuit.nbus,
                              nbr=numerical_circuit.nbr,
                              n_tr=numerical_circuit.ntr,
                              bus_names=numerical_circuit.bus_names,
                              branch_names=numerical_circuit.branch_names,
                              transformer_names=numerical_circuit.tr_names,
                              bus_types=numerical_circuit.bus_types,
                              calculation_inputs=calculation_inputs,
                              options=options,
                              dP=variation.dP,
                              return_dict=returns)

            pf_results, log = returns[0]
            results.logger += log

            # add the power flow results
            if v == 0:
                results.default_pf_results = pf_results
            else:
                results.add_results_at(v - 1, pf_results, variation)

            if prog_func is not None:
                p = (v + 1) / nvar * 100.0
                prog_func(p)

            if self.__cancel__:
                break

        return results

    def ptdf_multi_treading(self, circuit: MultiCircuit, options: PowerFlowOptions, group_mode: PtdfGroupMode,
                            power_amount, text_func=None, prog_func=None):
        """
        Power Transfer Distribution Factors analysis
        :param circuit: MultiCircuit instance
        :param options: power flow options
        :param group_mode: ptdf grouping mode
        :param power_amount: amount o power to vary in MW
        :param text_func:
        :param prog_func
        :return:
        """

        if text_func is not None:
            text_func('Compiling...')

        # compile to arrays
        # numerical_circuit = circuit.compile_snapshot()
        # calculation_inputs = numerical_circuit.compute(apply_temperature=options.apply_temperature_correction,
        #                                                branch_tolerance_mode=options.branch_impedance_tolerance_mode,
        #                                                ignore_single_node_islands=options.ignore_single_node_islands)

        numerical_circuit = compile_snapshot_circuit(circuit=circuit,
                                                     apply_temperature=options.apply_temperature_correction,
                                                     branch_tolerance_mode=options.branch_impedance_tolerance_mode,
                                                     opf_results=self.opf_results)

        calculation_inputs = numerical_circuit.split_into_islands(ignore_single_node_islands=options.ignore_single_node_islands)

        # compute the variations
        delta_of_power_variations = get_ptdf_variations(circuit=circuit,
                                                        numerical_circuit=numerical_circuit,
                                                        group_mode=group_mode,
                                                        power_amount=power_amount)

        # declare the PTDF results
        results = PTDFResults(n_variations=len(delta_of_power_variations) - 1,
                              n_br=numerical_circuit.nbr,
                              n_bus=numerical_circuit.nbus,
                              br_names=numerical_circuit.branch_names,
                              bus_names=numerical_circuit.bus_names,
                              bus_types=numerical_circuit.bus_types)

        if text_func is not None:
            text_func('Running PTDF...')

        jobs = list()
        n_cores = multiprocessing.cpu_count()
        manager = multiprocessing.Manager()
        return_dict = manager.dict()

        # for v, variation in enumerate(delta_of_power_variations):
        v = 0
        nvar = len(delta_of_power_variations)
        while v < nvar:

            k = 0

            # launch only n_cores jobs at the time
            while k < n_cores + 2 and (v + k) < nvar:
                # run power flow at the circuit
                p = multiprocessing.Process(target=power_flow_worker, args=(v,
                                                                            numerical_circuit.nbus,
                                                                            numerical_circuit.nbr,
                                                                            numerical_circuit.ntr,
                                                                            numerical_circuit.bus_names,
                                                                            numerical_circuit.branch_names,
                                                                            numerical_circuit.tr_names,
                                                                            numerical_circuit.bus_types,
                                                                            calculation_inputs,
                                                                            options,
                                                                            delta_of_power_variations[v].dP,
                                                                            return_dict))
                jobs.append(p)
                p.start()
                v += 1
                k += 1

                if self.__cancel__:
                    break

            # wait for all jobs to complete
            for process_ in jobs:
                process_.join()

            # emit the progress
            if prog_func is not None:
                p = (v + 1) / nvar * 100.0
                prog_func(p)

            if self.__cancel__:
                break

        if text_func is not None:
            text_func('Collecting results...')

        # gather the results
        if not self.__cancel__:
            for v in range(nvar):
                pf_results, log = return_dict[v]
                results.logger += log
                if v == 0:
                    results.default_pf_results = pf_results
                else:
                    results.add_results_at(v - 1, pf_results, delta_of_power_variations[v])

        return results

    def run(self):
        """
        Run thread
        """
        start = time.time()
        if self.options.use_multi_threading:

            self.results = self.ptdf_multi_treading(circuit=self.grid, options=self.pf_options,
                                                    group_mode=self.options.group_mode,
                                                    power_amount=self.options.power_increment,
                                                    text_func=self.progress_text.emit,
                                                    prog_func=self.progress_signal.emit)
        else:

            self.results = self.ptdf(circuit=self.grid, options=self.pf_options,
                                     group_mode=self.options.group_mode,
                                     power_amount=self.options.power_increment,
                                     text_func=self.progress_text.emit,
                                     prog_func=self.progress_signal.emit)

        if not self.__cancel__:
            self.results.consolidate()

        end = time.time()
        self.elapsed = end - start
        self.progress_text.emit('Done!')
        self.done_signal.emit()

    def get_steps(self):
        """
        Get variations list of strings
        """
        if self.results is not None:
            return [v.name for v in self.results.variations]
        else:
            return list()

    def cancel(self):
        self.__cancel__ = True


if __name__ == '__main__':

    from GridCal.Engine import FileOpen, SolverType

    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/Lynn 5 Bus pv.gridcal'
    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE39_1W.gridcal'
    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/grid_2_islands.xlsx'
    fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/1354 Pegase.xlsx'

    main_circuit = FileOpen(fname).open()

    pf_options = PowerFlowOptions(solver_type=SolverType.DC)
    options_ = PTDFOptions(group_mode=PtdfGroupMode.ByGenLoad, power_increment=10, use_multi_threading=False)
    simulation = PTDF(grid=main_circuit, options=options_, pf_options=pf_options)
    simulation.run()
    ptdf_df = simulation.results.get_flows_data_frame()

    print(ptdf_df)

    print()
    a = time.time()
    options_ = PTDFOptions(group_mode=PtdfGroupMode.ByGenLoad, power_increment=10, use_multi_threading=True)
    simulation = PTDF(grid=main_circuit, options=options_, pf_options=pf_options)
    simulation.run()
    ptdf_df = simulation.results.get_flows_data_frame()
    b = time.time()
    print(ptdf_df)
    print(b-a)





class CircuitMatrixResults:

    def __init__(self):
        """
        Class to store grid calculated values from Power flow ans State estimation
        """

        # node voltages vector (p.u.)
        self.V = None

        # node power injections (p.u.)
        self.Sbus = None

        # branch power injected at the from side (p.u.)
        self.Sf = None

        # branch power injected at the to side (p.u.)
        self.St = None

        # branch current injected at the from side (p.u.)
        self.If = None

        # branch current injected at the to side (p.u.)
        self.It = None

        # power flowing through the branch (p.u.)
        self.Sbranch = None

        # current flowing through the branch (p.u.)
        self.Ibranch = None

        # losses of the branch (p.u.)
        self.losses = None

        # did this solution converge?
        self.converged = False

        # power mismatch of this solution (p.u.)
        self.error = None




import numpy as np
from scipy.sparse import lil_matrix

from research.three_phase.Engine.general import *
from research.three_phase.Engine.circuit_matrix_results import *


class CircuitMatrixInputs:

    def __init__(self, n, m, n_phase):
        """
        Class to store the compiled circuit matrices and vectors
        :param n: number of nodes
        :param m: number of branches
        :param n_phase: circuit number of phases
        """

        # list of slack bus indices
        self.ref = list()

        # list of pq bus indices
        self.pq = list()

        # list of pv bus indices
        self.pv = list()

        # nodal voltage initial solution (p.u.)
        self.Vbus = np.ones((n * n_phase), dtype=np.complex)

        # nodal power injections (p.u.)
        self.Sbus = np.zeros((n * n_phase), dtype=np.complex)

        # nodal current injections (p.u.)
        self.Ibus = np.zeros((n * n_phase), dtype=np.complex)

        # exponential load model parameters
        self.P0 = np.zeros((n * n_phase), dtype=np.float)
        self.Q0 = np.zeros((n * n_phase), dtype=np.float)
        self.exp_p = np.zeros((n * n_phase), dtype=np.float)
        self.exp_q = np.zeros((n * n_phase), dtype=np.float)
        self.V0 = np.zeros((n * n_phase), dtype=np.float)

        # polynomial load model
        self.A = np.zeros((n * n_phase), dtype=np.float)
        self.B = np.zeros((n * n_phase), dtype=np.float)
        self.C = np.zeros((n * n_phase), dtype=np.float)

        # nodal admittances (p.u.)
        self.Ybus_sh = np.zeros((n * n_phase), dtype=np.complex)

        # nodal admittance matrix (p.u.)
        self.Ybus = lil_matrix((n * n_phase, n * n_phase), dtype=np.complex)  # sparse and complex in lil format
        self.Yseries = lil_matrix((n * n_phase, n * n_phase), dtype=np.complex)  # sparse and complex in lil format

        # self.Yf = lil_matrix((m * n_phase, n * n_phase), dtype=np.complex)
        # self.Yt = lil_matrix((m * n_phase, n * n_phase), dtype=np.complex)

        # branch-bus connectivity matrix
        self.Cf = lil_matrix((m * n_phase, n * n_phase), dtype=int)
        self.Ct = lil_matrix((m * n_phase, n * n_phase), dtype=int)

        # branch (lines and transformers) ratings in MVA (per wire)
        self.branch_rates = np.zeros((m * n_phase), dtype=int)

        # columns mask for Ybus
        self.mask_c = np.zeros((n * n_phase), dtype=int)

        # rows mask for Ybus, Sbus, Ibus, Vbus
        self.mask_r = np.zeros((n * n_phase), dtype=int)

        # branches mask
        self.mask_k = np.zeros((m * n_phase), dtype=int)

        # bus types array
        self.bus_types = np.zeros((n * n_phase), dtype=int)

        # buses mask
        self.bus_mask = None

    def consolidate(self):
        """
        Consolidate in-place the matrices and vectors so that
        the non-connected phases are removed (having non connected phases makes Ybus singular)
        """

        # find the indices of the connected phases
        idx_r = np.where(self.mask_r > 0)[0]
        idx_c = np.where(self.mask_c > 0)[0]
        idx_k = np.where(self.mask_k > 0)[0]
        self.bus_mask = self.mask_r + self.mask_c

        # reduce the matrices and vectors (eliminate the spurious phases)
        self.Vbus = self.Vbus[idx_r]
        self.Sbus = self.Sbus[idx_r]
        self.Ibus = self.Ibus[idx_r]

        # reduce the exponential load model parameters
        self.P0 = self.P0[idx_r]
        self.Q0 = self.Q0[idx_r]
        self.exp_p = self.exp_p[idx_r]
        self.exp_q = self.exp_q[idx_r]
        self.V0 = self.V0[idx_r]

        # reduce the polynomial load model
        self.A = self.A[idx_r]
        self.B = self.B[idx_r]
        self.C = self.C[idx_r]

        self.Ybus = self.Ybus[idx_r, :][:, idx_c]
        self.Yseries = self.Yseries[idx_r, :][:, idx_c]

        # self.Yf = self.Yf[idx_k, :][:, idx_c]
        # self.Yt = self.Yt[idx_k, :][:, idx_c]

        self.Cf = self.Cf[idx_k, :][:, idx_c]
        self.Ct = self.Ct[idx_k, :][:, idx_c]

        self.branch_rates = self.branch_rates[idx_k]

        # self.print('POST' + "-"*80)

        # determine the bus-type arrays
        self.bus_types = self.bus_types[idx_r]

        for k, tpe in enumerate(self.bus_types):

            # keep track of the node types
            if tpe == BusTypes.Ref.value:
                self.ref.append(k)

            elif tpe == BusTypes.PQ.value:
                self.pq.append(k)

            elif tpe == BusTypes.PV.value:
                self.pv.append(k)

    def compute_branch_results(self, V):
        """
        Compute the branch magnitudes from the voltages
        :param V: Voltage vector solution in p.u.
        :return: PowerFlowResults instance with all the grid magnitudes
        """

        # declare circuit results
        data = CircuitMatrixResults()

        # copy the voltage
        data.V = V

        # power at the slack nodes
        data.Sbus = self.Sbus.copy()
        data.Sbus[self.ref] = V[self.ref] * np.conj(self.Ybus[self.ref, :].dot(V))

        # Reactive power at the pv nodes: keep the original P injection and set the calculated reactive power
        Q = (V[self.pv] * np.conj(self.Ybus[self.pv, :].dot(V))).imag
        data.Sbus[self.pv] = self.Sbus[self.pv].real + 1j * Q

        # Branches current, loading, etc
        # data.If = self.Yf * V
        # data.It = self.Yt * V
        data.If = self.Cf * self.Ybus * V
        data.It = self.Ct * self.Ybus * V
        data.Sf = self.Cf * V * np.conj(data.If)
        data.St = self.Ct * V * np.conj(data.It)

        # Branch losses in MVA
        data.losses = (data.Sf + data.St)

        # Branch current in p.u.
        data.Ibranch = np.maximum(data.If, data.It)

        # Branch power in MVA
        data.Sbranch = np.maximum(data.Sf, data.St)

        # Branch loading in p.u.
        data.loading = data.Sbranch / (self.branch_rates + 1e-9)

        return data

    def print(self, msg=''):
        """

        :return:
        """
        print('\n\n', msg)
        print("Ybus:\n", self.Ybus.todense())
        print("mask_r:\n", self.mask_r)
        print("mask_c:\n", self.mask_c)

        # print("Yf:\n", self.Yf.todense())
        # print("Yt:\n", self.Yt.todense())
        # print("mask_f:\n", self.mask_f)
        # print("mask_t:\n", self.mask_t)
        print("mask_k:\n", self.mask_k)

        print("Cf:\n", self.Cf.todense())
        print("Ct:\n", self.Ct.todense())

        print("Vbus:\n", self.Vbus)
        print("Sbus:\n", self.Sbus)
        print("Ibus:\n", self.Ibus)

        # print("F:\n", self.F)
        # print("T:\n", self.T)
        print("Rates:\n", self.branch_rates)




from GridCal.Engine.Simulations.NK.n_minus_k_driver import *
from GridCal.Engine.Simulations.NK.n_minus_k_results import *
from GridCal.Engine.Simulations.NK.n_minus_k_ts_driver import *
from GridCal.Engine.Simulations.NK.n_minus_k_ts_results import *



# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.
import PySide2

from GridCal.Engine.Devices.editable_device import EditableDevice, DeviceType, GCProp
from PySide2 import QtCore


class Wire(EditableDevice):

    def __init__(self, name='', idtag=None, gmr=0.01, r=0.01, x=0.0, max_current=1):
        """
        Wire definition
        :param name: Name of the wire type
        :param gmr: Geometric Mean Radius (m)
        :param r: Resistance per unit length (Ohm / km)
        :param x: Reactance per unit length (Ohm / km)
        :param max_current: Maximum current of the conductor in (kA)

        """

        EditableDevice.__init__(self,
                                name=name,
                                idtag=idtag,
                                active=True,
                                device_type=DeviceType.WireDevice,
                                editable_headers={'name': GCProp('', str, "Name of the conductor"),
                                                  'idtag': GCProp('', str, 'Unique ID'),
                                                  'r': GCProp('Ohm/km', float, "resistance of the conductor"),
                                                  'x': GCProp('Ohm/km', float, "reactance of the conductor"),
                                                  'gmr': GCProp('m', float, "Geometric Mean Radius of the conductor"),
                                                  'max_current': GCProp('kA', float, "Maximum current of the conductor")
                                                  },
                                non_editable_attributes=list(),
                                properties_with_profile={})

        # self.wire_name = name
        self.r = r
        self.x = x
        self.gmr = gmr
        self.max_current = max_current

    def copy(self):
        """
        Copy of the wire
        :return:
        """
        # name='', idtag=None, gmr=0.01, r=0.01, x=0.0, max_current=1
        return Wire(name=self.name, gmr=self.gmr, r=self.r, x=self.x, max_current=self.max_current)


class WiresTable(QtCore.QAbstractTableModel):

    def __init__(self, parent=None):
        QtCore.QAbstractTableModel.__init__(self, parent)

        self.header = ['Name', 'R (Ohm/km)', 'GMR (m)']

        self.index_prop = {0: 'name', 1: 'r', 2: 'gmr'}

        self.converter = {0: str, 1: float, 2: float}

        self.editable = [True, True, True]

        self.wires = list()

    def add(self, wire: Wire):
        """
        Add wire
        :param wire:
        :return:
        """
        row = len(self.wires)
        self.beginInsertRows(QtCore.QModelIndex(), row, row)
        self.wires.append(wire)
        self.endInsertRows()

    def delete(self, index):
        """
        Delete wire
        :param index:
        :return:
        """
        row = len(self.wires)
        self.beginRemoveRows(QtCore.QModelIndex(), row - 1, row - 1)
        self.wires.pop(index)
        self.endRemoveRows()

    def is_used(self, name):
        """
        checks if the name is used
        """
        n = len(self.wires)
        for i in range(n-1, -1, -1):
            if self.wires[i].name == name:
                return True
        return False

    def flags(self, index):
        if self.editable[index.column()]:
            return QtCore.Qt.ItemIsEditable | QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable
        else:
            return QtCore.Qt.ItemIsEnabled

    def rowCount(self, parent=QtCore.QModelIndex()):
        return len(self.wires)

    def columnCount(self, parent=QtCore.QModelIndex()):
        return len(self.header)

    def parent(self, index=None):
        return QtCore.QModelIndex()

    def data(self, index, role=QtCore.Qt.DisplayRole):
        if index.isValid():
            if role == QtCore.Qt.DisplayRole:
                val = getattr(self.wires[index.row()], self.index_prop[index.column()])
                return str(val)
        return None

    def headerData(self, p_int, orientation, role):
        if role == QtCore.Qt.DisplayRole:
            if orientation == QtCore.Qt.Horizontal:
                return self.header[p_int]

    def setData(self, index, value, role=QtCore.Qt.DisplayRole):
        """
        Set data by simple editor (whatever text)
        :param index:
        :param value:
        :param role:
        """
        if self.editable[index.column()]:
            wire = self.wires[index.row()]
            attr = self.index_prop[index.column()]

            if attr == 'tower_name':
                if self.is_used(value):
                    pass
                else:
                    setattr(wire, attr, self.converter[index.column()](value))
            else:
                setattr(wire, attr, self.converter[index.column()](value))

        return True








import numpy as np

np.set_printoptions(precision=6, suppress=True, linewidth=320)
from numpy import where, zeros, ones, mod, conj, array, dot, complex128  # , complex256
from scipy.linalg import solve

from scipy.sparse.linalg import factorized
from scipy.sparse import issparse
from matplotlib import pyplot as plt

# Set the complex precision to use
complex_type = complex128


# @jit(cache=True)
def reduce_arrays(n_bus, Ymat, slack_indices, Vset, S, types):
    """
    Reduction of the circuit magnitudes.

    Args:
        n_bus: Number of buses of the circuit

        Ymat: Circuit admittance matrix

        slack_indices: Array of indices of the slack nodes

        Vset: Vector of voltages of those nodes where the voltage is controlled (AKA Slack and PV buses)

        S: Vector of power injections at all the nodes

        types: Vector of nde types

    Output:
        Yred: Reduced admittance matrix (Without the rows and columns belonging to slack buses)

        I: Matrix of currents (In practice only one slack bus is selected, hence it is a vector) injected by the slack buses

        Sred: Array of power injections of the buses that are not of type slack

        types_red: Array of types of the buses that are not of type slack

        non_slack_indices: Array of indices of the buses that are not of type slack
    """

    # Compose the list of buses indices excluding the indices of the slack buses
    # Compose the list of buses indices excluding the indices of the slack buses
    non_slack_indices = list(range(n_bus))
    for i in slack_indices[::-1]:
        non_slack_indices.pop(i)
    non_slack_indices = array(non_slack_indices)

    # Types of the non slack buses
    types_red = types[non_slack_indices]

    # now to have efficient arrays of coefficients
    map_idx = zeros(len(types_red), dtype=np.int)
    npq = 0
    npv = 0
    for i in range(len(types_red)):
        if types_red[i] == 1:  # PQ
            map_idx[i] = npq
            npq += 1
        elif types_red[i] == 2:  # PV
            map_idx[i] = npv
            npv += 1

    # Compose a reduced admittance matrix without the rows and columns that correspond to the slack buses
    Yred = Ymat[non_slack_indices, :][:, non_slack_indices]

    # matrix of the columns of the admittance matrix that correspond to the slack buses
    Yslack = Ymat[non_slack_indices, :][:, slack_indices]

    # vector of slack voltages (Complex)
    Vslack = Vset[slack_indices]

    # vector of currents being injected by the slack nodes (Matrix vector product)
    Iind = -1 * np.ndarray.flatten(array(Yslack.dot(Vslack)))
    # Iind /= abs(Iind)

    # Invert Yred: By storing into a factorization object we get a much better numerical performance
    # In the end, if Zred=Yred^-1,  Zred^-1 x I is the same as Yred_LU.solve(I)
    Zred = factorized(Yred)

    # Vind = Zred * Iind  # (not needed)

    # Vector of reduced power values (Non slack power injections)
    Sred = S[non_slack_indices]

    return Yred, Zred, Iind, Sred, Vslack, types_red, non_slack_indices, map_idx, npq, npv


# @jit(cache=True)
def I_n(n, nbus, Iind, S, Vset_abs2, C, H, R, W, types_red, map_idx):
    """
    Array of Intensity coefficients of order n
    @param n: Order of the coefficients
    @param nbus: Number of buses (not counting the slack buses)
    @param Iind: Vector of current injections (nbus elements)
    @param S: Vector of power injections (nbus elements)
    @param Vset_abs2: Vector with the voltage set points squared. (nbus elements)
    @param C: Voltage coefficients (order x node)
    @param H: R-X convolution structure (order x pv-node)
    @param R: Voltage convolution coefficients (order x pv-node)
    @param W: Inverse coefficients structure (order x pq-node)
    @param types_red: Types of the non slack buses (nbus elements)
    @param map_idx: node -> index map in the PQ or PV structures
    @return:
    """

    rhs = np.empty(nbus, dtype=complex_type)

    for k in range(nbus):
        kk = map_idx[k]
        if types_red[k] == 1:  # PQ
            rhs[k] = I_PQ(n, k, kk, Iind, S, W, map_idx)
        elif types_red[k] == 2:  # PV
            rhs[k] = I_PV(n, k, kk, S, Iind, Vset_abs2, C, H, R, map_idx)

    return rhs


# @jit(cache=True)
def I_PQ(n, k, kk, Iind, S, W, map_idx):
    """
    Value of the intensity coefficient of order n for a PQ node k
    @param n: Order of the coefficients
    @param k: Index of the bus
    @param Iind: Vector of reduced current injections (nbus elements)
    @param S: Vector of reduced power injections (nbus elements)
    @param W: Inverse coefficients structure (order x pq-node)
    @param map_idx: node -> index in the PQ-only structures map
    @return: complex value
    """

    if n == 0:
        return Iind[k]

    else:
        # kk = map_idx[k]
        return conj(S[k]) * W[n - 1, kk]


# @jit(cache=True)
def calc_W(n, k, kk, C, W, map_idx):
    """
    Calculation of the inverse coefficients W. (only applicable for PQ buses)
    @param n: Order of the coefficients
    @param k: Index of the bus
    @param C: Voltage coefficients (order x node)
    @param W: Inverse coefficients structure (order x pq-node)
    @param map_idx: node -> index in the PQ-only structures map
    @return: complex value
    """

    if n == 0:
        res = complex_type(1)

    else:
        # kk = map_idx[k]
        res = complex_type(0)
        for l in range(n):
            res -= W[l, kk] * conj(C[n - l, k])

    res /= conj(C[0, k])

    return res


# @jit(cache=True)
def I_PV(n, k, kk, Sred, Iind, Vset_abs2, C, H, R, map_idx):
    """
    Get the intensity coefficient of order n of the PV node k
    @param n: order of the coefficient
    @param k: index of the node
    @param Sred: reduced specified power array
    @param Iind: induced intensity in the non-slack nodes
    @param Vset_abs2: array of reduced set-voltages squared
    @param C: Voltage coefficients structure (order x node)
    @param H: R-X convolution (order x pv-node)
    @param R: Voltage convolution coefficients (order x pv-node)
    @param map_idx: node -> index in the PV-only structures map
    @return: complex value
    """

    if n == 0:
        return Iind[k]
    else:
        # kk = map_idx[k]
        rhs = (2.0 * Sred[k].real * C[n - 1, k] - H[n - 1, kk] + R[n - 1, kk] * conj(Iind[k])) / Vset_abs2[k]

    return rhs


# @jit(cache=True)
def calc_R(n, k, C):
    """
    Voltage convolution coefficient

    Args:
        n: Order of the coefficients

        k: Index of the bus

        C: Voltage coefficients (order x node)

    Output:
        Convolution coefficient of order n for the bus k
    """

    result = complex_type(0)
    for l in range(n + 1):
        result += C[l, k] * C[n - l, k]

    return result


# @jit(cache=True)
def calc_X(n, k, nbus, C, Yred):
    """
    Admittance weighted coefficient

    Args:
        n: Order of the coefficients

        k: Index of the bus

        nbus: Number of non-slack buses

        C: Voltage coefficients (Ncoeff x nbus elements)

        Yred: Reduced admittance matrix

    Output:
        Admittance weighted coefficient of order n for the bus k
    """
    result = complex_type(0)
    for i in range(nbus):
        if i != k:
            result += conj(Yred[k, i] * C[n, i])

    return result


# @jit(cache=True)
def calc_H(n, k, kk, X, R, C, Yred, Vset_abs2, map_idx):
    """
    Calculate the combined coefficient
    Args:
        n: Order of the coefficients

        k: Index of the bus

        X: Weighted coefficients (Ncoeff x nbus elements)

        R: Convolution coefficient (Ncoeff x nbus elements)

        C: Voltage coefficients (Ncoeff x nbus elements)

        Yred: Reduced admittance matrix

        Vset_abs2: Vector with the voltage set points squared. (nbus elements)

    Output:
        Returns the H coefficient of order n for the bus of index k
    """
    # kk = map_idx[k]
    #
    result = complex_type(0)
    for i in range(n + 1):
        result += X[n - i, kk] * R[i, kk]
    result += conj(Yred[k, k]) * C[n, k] * Vset_abs2[k]

    return result


# @jit(cache=True)
def epsilon(Sn, n, E):
    """
    Fast recursive Wynn's epsilon algorithm from:
        NONLINEAR SEQUENCE TRANSFORMATIONS FOR THE ACCELERATION OF CONVERGENCE
        AND THE SUMMATION OF DIVERGENT SERIES

        by Ernst Joachim Weniger
    """
    Zero = complex_type(0)
    One = complex_type(1)
    Tiny = np.finfo(complex_type).min
    Huge = np.finfo(complex_type).max

    E[n] = Sn

    if n == 0:
        estim = Sn
    else:
        AUX2 = Zero

        for j in range(n, 0, -1):  # range from n to 1 (both included)
            AUX1 = AUX2
            AUX2 = E[j - 1]
            DIFF = E[j] - AUX2

            if abs(DIFF) <= Tiny:
                E[j - 1] = Huge
            else:
                if DIFF == 0.0:
                    DIFF = Tiny
                E[j - 1] = AUX1 + One / DIFF

        if mod(n, 2) == 0:
            estim = E[0]
        else:
            estim = E[1]

    return estim, E


# @jit(cache=True)
def pade_approximation(n, d, an, s=1):
    """
    Computes the n/2 pade approximant of the series an at the approximation
    point s

    Arguments:
        an: coefficient series
        n:  order of the series
        d:  bus index
        s: point of approximation

    Returns:
        pade approximation at s
    """
    nn = int(n / 2)
    L = nn
    M = nn

    # formation of the linear system right hand side
    rhs = an[L + 1:L + M + 1, d]

    # formation of the coefficients matrix
    C = zeros((L, M), dtype=complex_type)
    for i in range(L):
        k = i + 1
        C[i, :] = an[L - M + k:L + k, d]

    # Obtaining of the b coefficients for orders greater than 0
    b = solve(C, -rhs)  # bn to b1
    b = r_[1, b[::-1]]  # b0 = 1

    # Obtaining of the coefficients 'a'
    a = zeros(L + 1, dtype=complex_type)
    a[0] = an[0, d]
    for i in range(L):
        val = complex_type(0)
        k = i + 1
        for j in range(k + 1):
            val += an[k - j, d] * b[j]
        a[i + 1] = val

    # evaluation of the function for the value 's'
    p = complex_type(0)
    q = complex_type(0)
    for i in range(L + 1):
        p += a[i] * s ** i
        q += b[i] * s ** i

    return p / q, a, b


# @jit(cache=True)
def update_bus_power(k, V, Y):
    """
    Computes the power for a PV or VD bus
    """
    return V[k] * conj(Y[k, :].dot(V))


# @jit(cache=True)
def update_all_powers(pv_idx_all, slack_idx_all, V, Y, Sbus):
    """
    Computes the power for all the PV buses and VD buses
    """
    S = Sbus.copy().astype(complex_type)

    # update reactive power for all PV buses
    for k in pv_idx_all:
        Q = update_bus_power(k, V, Y)[0].imag
        S[k] = Sbus[k].real + 1j * Q

    for k in slack_idx_all:
        S[k] = update_bus_power(k, V, Y)[0]

    return S


# @jit(cache=True)
def calc_error(admittances, V, powerInjections):
    """
    Calculates the power error for all the buses
    """
    v_mat = np.diag(V)
    vy_mat = conj(admittances.dot(V))
    return powerInjections - dot(v_mat, vy_mat)


# @jit(cache=True)
def helmz(admittances, slackIndices, maxcoefficientCount, powerInjections, voltageSetPoints, types,
          eps=1e-3, usePade=True, inherited_pv=None):
    """

    Args:
        admittances: Circuit complete admittance matrix

        slackIndices: Indices of the slack buses (although most likely only one works)

        coefficientCount: Number of voltage coefficients to evaluate (Must be an odd number)

        powerInjections: Array of power injections matching the admittance matrix size

        voltageSetPoints: Array of voltage set points matching the admittance matrix size

        types: Array of bus types matching the admittance matrix size. types: {1-> PQ, 2-> PV, 3-> Slack}

    Output:
        Voltages vector
    """

    # The routines in this script are meant to handle sparse matrices, hence non-sparse ones are not allowed
    assert (issparse(admittances))

    # get the admittance matrix size AKA number of nodes
    n_original = np.shape(admittances)[0]

    # get array with the PV buses indices
    pq_idx_all = where(types == 1)[0]
    pv_idx_all = where(types == 2)[0]
    # reduce the admittance matrix to omit the slack buses
    Yred, Zred, Iind, Sred, Vslack, types_red, non_slack_indices, \
    map_idx, npq, npv = reduce_arrays(n_bus=n_original, Ymat=admittances.copy(),
                                      slack_indices=array(slackIndices, dtype=int),
                                      Vset=voltageSetPoints, S=powerInjections,
                                      types=types)

    # get the new dimension AKA number of nodes minus the slack nodes
    nbus = len(non_slack_indices)

    # declare the matrix of coefficients that will lead to the voltage computation
    C = zeros((0, nbus), dtype=complex_type)

    # auxiliary array for the epsilon algorithm
    E = zeros((0, nbus), dtype=complex_type)

    # Declare the inverse coefficients vector
    # (it is actually a matrix; a vector of coefficients per coefficient order)
    W = zeros((0, npq), dtype=complex_type)

    # Admittance weighted coefficients
    X = zeros((0, npv), dtype=complex_type)

    # Convolution coefficients
    R = zeros((0, npv), dtype=complex_type)

    # Convolution coefficients
    H = zeros((0, npv), dtype=complex_type)

    # Squared values of the voltage module for the buses that are not of slack type
    Vset_abs2 = (abs(voltageSetPoints) ** 2)[non_slack_indices]

    # progressive calculation of coefficients
    n = 0
    converged = False
    inside_precission = True
    errors = list()
    errors_PV_P = list()
    errors_PV_Q = list()
    errors_PQ_P = list()
    errors_PQ_Q = list()
    voltages = list()
    Sn = zeros(nbus, dtype=complex_type)
    Vred = zeros(nbus, dtype=complex_type)
    Vred_last = zeros(nbus, dtype=complex_type)
    best_V = None
    last_err = 1e100

    # Declare the vector of all the voltages
    voltages_vector = zeros(n_original, dtype=complex_type)
    voltages_vector[slackIndices] = Vslack

    error = list()

    while n <= maxcoefficientCount and not converged and inside_precission:

        # Reserve coefficients memory space
        C = np.vstack((C, np.zeros((1, nbus), dtype=complex_type)))
        E = np.vstack((E, np.zeros((1, nbus), dtype=complex_type)))
        W = np.vstack((W, np.zeros((1, npq), dtype=complex_type)))
        R = np.vstack((R, np.zeros((1, npv), dtype=complex_type)))
        X = np.vstack((X, np.zeros((1, npv), dtype=complex_type)))
        H = np.vstack((H, np.zeros((1, npv), dtype=complex_type)))

        # get the system independent term to solve the coefficients
        I = I_n(n, nbus, Iind, Sred, Vset_abs2, C, H, R, W, types_red, map_idx)

        # C[n, :] = np.dot(Zred, I)
        C[n, :] = Zred(I)

        # check NaN's
        if not np.isnan(np.sum(C[n, :])):

            E[n, :] = C[n, :]
            Sn += C[n, :]

            # Update the auxiliary coefficients (Inverse, Squared and Admittance weighted)
            for j in range(nbus):
                if types_red[j] == 1:  # PQ
                    kk = map_idx[j]  # actual index in the coefficients structure
                    W[n, kk] = calc_W(n, j, kk, C, W, map_idx)
                elif types_red[j] == 2:  # PV
                    kk = map_idx[j]  # actual index in the coefficients structures
                    R[n, kk] = calc_R(n, j, C)
                    X[n, kk] = calc_X(n, j, nbus, C, Yred)
                    H[n, kk] = calc_H(n, j, kk, X, R, C, Yred, Vset_abs2, map_idx)

                # calculate the voltages
                if usePade:
                    if mod(n, 2) == 0 and n > 0:
                        Vred[j], _, _ = pade_approximation(n, j, C)

                else:
                    Vred[j], E[:, j] = epsilon(Sn[j], n, E[:, j])

                if np.isnan(Vred[j]):
                    print('Maximum precession reached at ', n)
                    Vred = Vred_last
                    inside_precission = False
                    break

            # check NaN's
            if np.isnan(np.sum(R[n, :]) + np.sum(X[n, :]) + np.sum(W[n, :]) + np.sum(H[n, :])):
                inside_precission = False
                print('Maximum precession reached at ', n)

            Vred_last = Vred.copy()
        else:
            inside_precission = False
            print('Maximum precession reached at ', n)

        # Assign the non-slack voltages
        voltages_vector[non_slack_indices] = Vred

        # Compose the voltage values from the coefficient series
        # voltages_vector = recompose_voltages(n, nbus, n_original, slackIndices, non_slack_indices, Vslack, C)
        voltages.append(voltages_vector.copy())

        # Calculate the missing power values (Q for PV buses and P, Q for slack buses)
        S = update_all_powers(pv_idx_all, slackIndices, voltages_vector, admittances, powerInjections)

        # Calculate the error and check the convergence
        mis = calc_error(admittances, voltages_vector, S)

        # check for convergence
        # Calculate the error and check the convergence
        Scalc = voltages_vector * conj(admittances * voltages_vector)

        # complex power mismatch
        power_mismatch = Scalc - S

        # concatenate error by type
        mismatch = r_[power_mismatch[pv_idx_all].real, power_mismatch[pq_idx_all].real, power_mismatch[pq_idx_all].imag]

        # check for convergence
        normF = linalg.norm(mismatch, Inf)
        errors.append(normF)

        if normF < last_err:
            last_err = normF
            best_V = voltages_vector

        converged = (normF < eps)

        if npv > 0:
            a = linalg.norm(mismatch[pv_idx_all].real, Inf)
        else:
            a = 0
        b = linalg.norm(mismatch[pq_idx_all].real, Inf)
        c = linalg.norm(mismatch[pq_idx_all].imag, Inf)
        error.append([a, b, c])

        n += 1

    errors_lst = [array(errors), array(errors_PV_P), array(errors_PV_Q), array(errors_PQ_P), array(errors_PQ_Q)]

    err_df = pd.DataFrame(array(error), columns=['PV_real', 'PQ_real', 'PQ_imag'])
    err_df.plot(logy=True)

    return best_V, C, W, X, R, H, Yred, errors_lst, converged, last_err, S, voltages


def bifurcation_point(C):
    """
    Computes the bifurcation point
    @param C:
    @return:
    """
    npoints = 100
    order_num, bus_num = np.shape(C)
    # V(S) = P(S)/Q(S)
    V = zeros((npoints, bus_num), dtype=complex_type)
    L = zeros((npoints, bus_num))
    for k in range(bus_num):
        _, p, q = pade_approximation(order_num, k, C)
        # print(k, 'P:', p)
        # print(k, 'Q:', q)
        asint = np.roots(q[::-1])
        asint = np.sort(abs(asint))
        asint = asint[asint > 2]
        print('Asymptotes', asint)
        # print('Asymptote:', asint[0])

        bpoint = asint[0]
        lmda = np.linspace(1, bpoint, npoints + 1)[:-1]
        # lmda = np.linspace(1, 100, npoints)

        pval = np.polyval(p[::-1], lmda)
        qval = np.polyval(q[::-1], lmda)

        V[:, k] = pval / qval
        L[:, k] = lmda

    return V, L


# @jit(cache=True)
def continued_fraction(n, k, C):
    """
    Pade approximation using continued fractions
    @param C: Voltage coefficients structure
    @param n: number of orders (rows of C)
    @param k: bus index (column of C)
    @return:
    """
    temp = complex_type(0)
    a = complex_type(1)

    for ni in range(n - 1, 0, -1):
        b = C[ni, k]

        # print(a, b, a/b)
        temp = b / (a + temp)

    return C[0, k] + temp


def print_coeffs(C, W, R, X, H):
    print('C')
    print(C.copy())
    print('W')
    print(W.copy())
    print('R')
    print(R.copy())
    print('X')
    print(X.copy())
    print('H')
    print(H.copy())


def print_polar(x):
    n = len(x)
    for i in range(n):
        print(abs(x[i]), ' < ', np.angle(x[i], True), 'd')


def to_latex(a, label='A'):
    import sys
    sys.stdout.write('\[ '
                     + label
                     + ' = \\left| \\begin{array}{'
                     + ('c' * a.shape[1])
                     + '}\n')
    for r in a:
        sys.stdout.write(str(r[0]))
        for c in r[1:]:
            sys.stdout.write(' & ' + str(c))
        sys.stdout.write('\\\\\n')
    sys.stdout.write('\\end{array} \\right| \]\n')


def plot_full_convergence(err, powerInjections, S, title, ext='.eps', save=True):
    plt.figure()

    titles = ['Maximum mismatch', 'PV nodes P mismatch', 'PV nodes Q mismatch', 'PQ nodes P mismatch',
              'PQ nodes Q mismatch']
    idx = [0, 1, 3, 4]
    k = 0
    for i in idx:
        k += 1
        plt.subplot(2, 2, k)
        if not np.all(err[i] == 0):
            plt.plot(err[i])
            plt.title(titles[i])
            try:
                plt.yscale('log')
            except:
                print()

    fig = plt.gcf()
    fig.suptitle(title, fontsize=14)

    # if save:
    #     plt.savefig(replace_non_ascii(title).replace(" ", "_") + ext)
    # else:
    #     plt.show()


if __name__ == "__main__":
    from GridCal.Engine.calculation_engine import *
    from matplotlib import pyplot as plt

    grid = MultiCircuit()
    grid.load_file('lynn5buspv.xlsx')
    # grid.load_file('IEEE30.xlsx')

    grid.compile()

    circuit = grid.circuits[0]

    print('\nYbus:\n', circuit.power_flow_input.Ybus.todense())
    print('\nYseries:\n', circuit.power_flow_input.Yseries.todense())
    print('\nYshunt:\n', circuit.power_flow_input.Yshunt)
    print('\nSbus:\n', circuit.power_flow_input.Sbus)
    print('\nIbus:\n', circuit.power_flow_input.Ibus)
    print('\nVbus:\n', circuit.power_flow_input.Vbus)
    print('\ntypes:\n', circuit.power_flow_input.types)
    print('\npq:\n', circuit.power_flow_input.pq)
    print('\npv:\n', circuit.power_flow_input.pv)
    print('\nvd:\n', circuit.power_flow_input.ref)

    import time
    print('HELM-Z')
    start_time = time.time()
    cmax = 25
    V1, C, W, X, R, H, Yred, err, converged_, \
    best_err, S, Vlst = helmz(admittances=circuit.power_flow_input.Ybus,
                              slackIndices=circuit.power_flow_input.ref,
                              maxcoefficientCount=cmax,
                              powerInjections=circuit.power_flow_input.Sbus,
                              voltageSetPoints=circuit.power_flow_input.Vbus,
                              types=circuit.power_flow_input.types,
                              eps=1e-9,
                              usePade=True,
                              inherited_pv=None)

    print("--- %s seconds ---" % (time.time() - start_time))
    # print_coeffs(C, W, R, X, H)

    print('V module:\t', abs(V1))
    print('V angle: \t', np.angle(V1))
    print('error: \t', best_err)

    # check the HELM solution: v against the NR power flow
    print('\nNR')
    options = PowerFlowOptions(SolverType.NR, verbose=False, robust=False, tolerance=1e-9)
    power_flow = PowerFlow(grid, options)

    start_time = time.time()
    power_flow.run()
    print("--- %s seconds ---" % (time.time() - start_time))
    vnr = circuit.power_flow_results.voltage

    print('V module:\t', abs(vnr))
    print('V angle: \t', np.angle(vnr))
    print('error: \t', circuit.power_flow_results.error)

    # check
    print('\ndiff:\t', V1 - vnr)

    plt.show()



# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pandas as pd

from GridCal.Engine.basic_structures import Logger
from GridCal.Engine.Simulations.result_types import ResultTypes
from GridCal.Engine.Simulations.PowerFlow.power_flow_driver import PowerFlowResults
from GridCal.Engine.Simulations.results_model import ResultsModel


class PTDFVariation:

    def __init__(self, name, n, original_power):
        """
        PTDF variation
        :param name: name of the variation
        :param n: number of buses
        :param original_power: power failed in MW
        """
        self.name = name

        # vector that was actually subtracted to Sbus
        self.dP = np.zeros(n)

        # original amount of power failed
        self.original_power = original_power


class PTDFResults:

    def __init__(self, n_variations=0, n_br=0, n_bus=0, br_names=(), bus_names=(), bus_types=()):
        """
        Number of variations
        :param n_variations:
        :param n_br: number of branches:
        """

        self.name = 'PTDF'

        # number of variations
        self.n_variations = n_variations

        # number of branches
        self.n_br = n_br

        self.n_bus = n_bus

        # names of the branches
        self.br_names = br_names

        self.bus_names = bus_names

        self.bus_types = bus_types

        # default power flow results
        self.default_pf_results = None

        # results of the variation
        self.pf_results = [None] * n_variations  # type: List[PowerFlowResults]

        # definition of the variation
        self.variations = [None] * n_variations  # type: List[PTDFVariation]

        self.logger = Logger()

        self.flows_sensitivity_matrix = None
        self.voltage_sensitivity_matrix = None

        self.available_results = [ResultTypes.PTDFBranchesSensitivity,
                                  ResultTypes.PTDFBusVoltageSensitivity]

    def add_results_at(self, i, results: PowerFlowResults, variation: PTDFVariation):
        """
        Add the results
        :param i: variation index
        :param results: PowerFlowResults instance
        :param variation: PTDFVariation instance
        :return: None
        """
        # store the results
        self.pf_results[i] = results
        self.variations[i] = variation

    def get_branch_sensitivity_at(self, i):
        """
        get Branch sensitivities
        :param i: variation index
        :return: array of sensitivities from -1 to 1
        """
        delta = (self.pf_results[i].Sf.real - self.default_pf_results.Sf.real)
        # self.variations[i].original_power is the power increment
        return delta / (self.variations[i].original_power + 1e-20)

    def get_voltage_sensitivity_at(self, i):
        """
        get Branch sensitivities
        :param i: variation index
        :return: array of sensitivities from -1 to 1
        """
        v0 = np.abs(self.default_pf_results.voltage)
        delta = (np.abs(self.pf_results[i].voltage) - v0)
        return delta / (self.variations[i].original_power + 1e-20)

    def get_var_names(self):
        """
        Get variation names
        :return:
        """
        return [v.name for v in self.variations]

    def consolidate(self):
        """
        Consolidate results in matrix
        :return:
        """
        self.flows_sensitivity_matrix = np.zeros((self.n_variations, self.n_br))
        for i in range(self.n_variations):
            self.flows_sensitivity_matrix[i, :] = self.get_branch_sensitivity_at(i)

        self.voltage_sensitivity_matrix = np.zeros((self.n_variations, self.n_bus))
        for i in range(self.n_variations):
            self.voltage_sensitivity_matrix[i, :] = self.get_voltage_sensitivity_at(i)

    def get_flows_data_frame(self):
        """
        Get Pandas DataFrame with the results
        :return: pandas DataFrame
        """

        if self.flows_sensitivity_matrix is None:
            self.consolidate()

        var_names = self.get_var_names()
        df = pd.DataFrame(data=self.flows_sensitivity_matrix.T, index=self.br_names, columns=var_names).fillna(0)

        return df

    def mdl(self, result_type: ResultTypes) -> ResultsModel:
        """
        Plot the results.

        Arguments:

            **result_type**: ResultTypes

            **indices**: Indices f the array to plot (indices of the elements)

            **names**: Names of the elements

        Returns:

            DataFrame
        """

        if result_type == ResultTypes.PTDFBranchesSensitivity:
            labels = self.br_names
            y = self.flows_sensitivity_matrix
            y_label = '(p.u.)'
            title = 'Branches sensitivity'

        elif result_type == ResultTypes.PTDFBusVoltageSensitivity:
            labels = self.bus_names
            y = self.voltage_sensitivity_matrix
            y_label = '(p.u.)'
            title = 'Buses voltage sensitivity'

        else:
            labels = []
            y = np.zeros(0)
            y_label = ''
            title = ''

        # assemble model
        mdl = ResultsModel(data=y, index=self.get_var_names(), columns=labels, title=title,
                           ylabel=y_label, units=y_label)
        return mdl



# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.
import numpy as np
import scipy.sparse as sp
import GridCal.Engine.Core.topology as tp


class LinesData:

    def __init__(self, nline, nbus, ntime=1):
        """

        :param nline:
        :param nbus:
        """
        self.nline = nline
        self.ntime = ntime

        self.line_names = np.zeros(nline, dtype=object)
        self.line_R = np.zeros(nline, dtype=float)
        self.line_X = np.zeros(nline, dtype=float)
        self.line_B = np.zeros(nline, dtype=float)

        self.C_line_bus = sp.lil_matrix((nline, nbus), dtype=int)  # this ons is just for splitting islands

    def slice(self, elm_idx, bus_idx, time_idx=None):
        """

        :param elm_idx:
        :param bus_idx:
        :return:
        """

        data = LinesData(nline=len(elm_idx), nbus=len(bus_idx))

        data.line_names = self.line_names[elm_idx]
        data.line_R = self.line_R[elm_idx]
        data.line_X = self.line_X[elm_idx]
        data.line_B = self.line_B[elm_idx]
        data.C_line_bus = self.C_line_bus[np.ix_(elm_idx, bus_idx)]

        return data

    def get_island(self, bus_idx):
        """
        Get the elements of the island given the bus indices
        :param bus_idx: list of bus indices
        :return: list of line indices of the island
        """
        return tp.get_elements_of_the_island(self.C_line_bus, bus_idx)

    def __len__(self):
        return self.nline




# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.


import numba as nb
import numpy as np
from scipy.sparse import csc_matrix


# @nb.njit("c16[:](i8, i4[:], i4[:], c16[:], c16[:], c16[:], i8)", parallel=True)
@nb.njit()
def calc_power_csr_numba(n, Yp, Yj, Yx, V, I, n_par=500):
    """
    Compute the power vector from the CSR admittance matrix
    :param m: number of rows
    :param n: number of columns
    :param Yp: pointers
    :param Yj: indices
    :param Yx: data
    :param V: vector x (n)
    :param I
    :param n_par: Number upon which the computation is done in parallel
    :return: vector y (m)
    """

    assert n == V.shape[0]
    S = np.zeros(n, dtype=nb.complex128)

    if n < n_par:
        # serial version
        for i in range(n):  # for every row
            s = complex(0, 0)
            for p in range(Yp[i], Yp[i+1]):  # for every column
                s += Yx[p] * V[Yj[p]]
            S[i] = V[i] * np.conj(s - I[i])
    else:
        # parallel version
        for i in nb.prange(n):  # for every row
            s = complex(0, 0)
            for p in range(Yp[i], Yp[i+1]):  # for every column
                s += Yx[p] * V[Yj[p]]
            S[i] = V[i] * np.conj(s - I[i])
    return S


# @nb.njit("Tuple((i4[:], i4[:], c16[:]))(i8, c16[:])")
@nb.njit()
def csc_diagonal_from_array(m, array):
    """

    :param m:
    :param array:
    :return:
    """
    indptr = np.empty(m + 1, dtype=nb.int32)
    indices = np.empty(m, dtype=nb.int32)
    data = np.empty(m, dtype=nb.complex128)
    for i in range(m):
        indptr[i] = i
        indices[i] = i
        data[i] = array[i]
    indptr[m] = m

    return indices, indptr, data


def diag(x):
    m = x.shape[0]
    indices, indptr, data = csc_diagonal_from_array(m, x)
    return csc_matrix((data, indices, indptr), shape=(m, m))



# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.
import time
import datetime
import numpy as np
from itertools import combinations
from PySide2.QtCore import QThread, Signal

from GridCal.Engine.basic_structures import Logger
from GridCal.Engine.Core.multi_circuit import MultiCircuit
from GridCal.Engine.Core.snapshot_pf_data import compile_snapshot_circuit
from GridCal.Engine.Simulations.NK.n_minus_k_results import NMinusKResults
from GridCal.Engine.Simulations.LinearFactors.analytic_ptdf import LinearAnalysis


def enumerate_states_n_k(m, k=1):
    """
    Enumerates the states to produce the so called N-k failures
    :param m: number of branches
    :param k: failure level
    :return: binary array (number of states, m)
    """

    # num = int(math.factorial(k) / math.factorial(m-k))
    states = list()
    indices = list()
    arr = np.ones(m, dtype=int).tolist()

    idx = list(range(m))
    for k1 in range(k + 1):
        for failed in combinations(idx, k1):
            indices.append(failed)
            arr2 = arr.copy()
            for j in failed:
                arr2[j] = 0
            states.append(arr2)

    return np.array(states), indices


class NMinusKOptions:

    def __init__(self, distributed_slack=True, correct_values=True):

        self.distributed_slack = distributed_slack

        self.correct_values = correct_values


class NMinusK(QThread):
    progress_signal = Signal(float)
    progress_text = Signal(str)
    done_signal = Signal()
    name = 'N-1/OTDF'

    def __init__(self, grid: MultiCircuit, options: NMinusKOptions):
        """
        N - k class constructor
        @param grid: MultiCircuit Object
        @param options: N-k options
        @:param pf_options: power flow options
        """
        QThread.__init__(self)

        # Grid to run
        self.grid = grid

        # Options to use
        self.options = options

        # N-K results
        self.results = NMinusKResults(n=0, m=0,
                                      bus_names=(),
                                      branch_names=(),
                                      bus_types=())

        self.numerical_circuit = None

        # set cancel state
        self.__cancel__ = False

        self.logger = Logger()

        self.elapsed = 0.0

        self.branch_names = list()

    def get_steps(self):
        """
        Get variations list of strings
        """
        if self.results is not None:
            return [v for v in self.branch_names]
        else:
            return list()

    def n_minus_k(self):
        """
        Run N-1 simulation in series
        :return: returns the results
        """

        self.progress_text.emit("Filtering elements by voltage")

        self.numerical_circuit = compile_snapshot_circuit(self.grid)

        results = NMinusKResults(m=self.numerical_circuit.nbr,
                                 n=self.numerical_circuit.nbus,
                                 branch_names=self.numerical_circuit.branch_names,
                                 bus_names=self.numerical_circuit.bus_names,
                                 bus_types=self.numerical_circuit.bus_types)

        self.progress_text.emit('Analyzing outage distribution factors...')
        linear_analysis = LinearAnalysis(grid=self.grid,
                                         distributed_slack=self.options.distributed_slack,
                                         correct_values=self.options.correct_values)
        linear_analysis.run()

        Pbus = self.numerical_circuit.get_injections(False).real[:, 0]
        PTDF = linear_analysis.results.PTDF
        LODF = linear_analysis.results.LODF

        # compute the branch flows in "n"
        flows_n = np.dot(PTDF, Pbus)

        self.progress_text.emit('Computing flows...')
        nl = self.numerical_circuit.nbr
        for c in range(nl):  # branch that fails (contingency)

            results.Sf[:, c] = flows_n[:] + LODF[:, c] * flows_n[c]
            results.loading[:, c] = results.Sf[:, c] / (self.numerical_circuit.branch_rates + 1e-9)

            results.S[c, :] = Pbus

            self.progress_signal.emit((c+1) / nl * 100)

        results.otdf = LODF

        return results

    def run(self):
        """

        :return:
        """
        start = time.time()
        self.results = self.n_minus_k()

        end = time.time()
        self.elapsed = end - start
        self.progress_text.emit('Done!')
        self.done_signal.emit()

    def cancel(self):
        self.__cancel__ = True


if __name__ == '__main__':
    import os
    import pandas as pd
    from GridCal.Engine import FileOpen, SolverType,PowerFlowOptions

    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/Lynn 5 Bus pv.gridcal'
    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE39_1W.gridcal'
    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/grid_2_islands.xlsx'
    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/2869 Pegase.gridcal'
    fname = os.path.join('..', '..', '..', '..', '..', 'Grids_and_profiles', 'grids', 'IEEE 30 Bus with storage.xlsx')
    # fname = os.path.join('..', '..', '..', '..', '..', 'Grids_and_profiles', 'grids', '2869 Pegase.gridcal')

    main_circuit = FileOpen(fname).open()

    options_ = NMinusKOptions()
    simulation = NMinusK(grid=main_circuit, options=options_)
    simulation.run()

    otdf_ = simulation.get_otdf()

    # save the result
    br_names = [b.name for b in main_circuit.branches]
    br_names2 = ['#' + b.name for b in main_circuit.branches]
    w = pd.ExcelWriter('OTDF IEEE30.xlsx')
    pd.DataFrame(data=simulation.results.Sf.real,
                 columns=br_names,
                 index=['base'] + br_names2).to_excel(w, sheet_name='branch power')
    pd.DataFrame(data=otdf_,
                 columns=br_names,
                 index=br_names2).to_excel(w, sheet_name='OTDF')
    w.save()


import numpy as np

from research.three_phase.Engine.general import *


class Load:
    """
    Load class template
    """
    def get_values(self):
        pass


class LoadSIY(Load):

    def __init__(self, name, S, I, Y,
                 conn=Phases.ABC,
                 conn_type=Connection.Delta):
        """
        Load constructor for the S-I-Y models (variant of the Z-I-P model)
        :param name: name of the element
        :param S: vector of complex power in MW + jMVA
        :param I: vector of complex current in MW + jMVA
        :param Y: vector of complex admittance in MW + jMVA
        :param conn: phases of connection
        :param conn_type: type of connection (Wye or Delta)

        If the connection is Delta the vectors contain [AB, BC, CA]
        If the connection is wye the vectors contain [AN, BN, CN]
        """

        self.name = name

        # type of the load object
        self.type = LoadTypes.ZIP

        # power in MW + jMVA
        self.S = S

        # current in MW + jMVA
        self.I = I

        # admittance in MW + jMVA
        self.Y = Y

        # connection type
        self.conn_type = conn_type

        # vector of the used phases
        self.phases = np.array(conn)

        # number of phases
        self.number_of_phases = len(self.phases)

        # if self.conn_type != Connection.PositiveSequence:
        #     if len(self.S) != len(self.phases):
        #         raise Exception(self.name + ': The power array does not match the specified phases')
        #
        #     if len(self.I) != len(self.phases):
        #         raise Exception(self.name + ': The current array does not match the specified phases')
        #
        #     if len(self.Y) != len(self.phases):
        #         raise Exception(self.name + ': The admittance array does not match the specified phases')
        # else:
        #     if self.number_of_phases != 1:
        #         raise Exception(self.name + ': Positive sequence requires to have exactly one phase: 0')

    def get_values(self):
        """
        Returns the wye-connected power, current and admittance
        :return: Wye-connected power
        """
        if self.number_of_phases == 3:
            if self.conn_type == Connection.Wye:
                return self.S, self.I, self.Y
            else:
                return delta_to_wye(self.S), delta_to_wye(self.I), delta_to_wye(self.Y)
        else:
            return self.S, self.I, self.Y


class LoadExp(Load):

    def __init__(self, name, P0, Q0, V0, exp_p, exp_q,
                 conn=Phases.ABC,
                 conn_type=Connection.Delta):
        """
        Load constructor for the exponential load model
        P = P0 * (V / V0)^exp_p
        Q = Q0 * (V / V0)^exp_q
        :param name: name of the element
        :param P0: Per unit active power at the per unit voltage V0  (vector of values per phase)
        :param Q0: Per unit reactive power at the per unit voltage V0 (vector of values per phase)
        :param V0: Nominal voltage in per unit [typical value: 1] (vector of values per phase)
        :param exp_p: active power exponent [typical values 0.4~0.8] (vector of values per phase)
        :param exp_q: reactive power exponent [typical values 2~3](vector of values per phase)
        :param conn: phases of connection
        :param conn_type: type of connection (Wye or Delta)

        If the connection is Delta the vectors contain [AB, BC, CA]
        If the connection is wye the vectors contain [AN, BN, CN]
        """

        self.name = name

        # type of the load object
        self.type = LoadTypes.Exponential

        # load model parameters
        self.P0 = P0
        self.Q0 = Q0
        self.V0 = V0
        self.exp_p = exp_p
        self.exp_q = exp_q

        # connection type
        self.conn_type = conn_type

        # vector of the used phases
        self.phases = np.array(conn)

        # number of phases
        self.number_of_phases = len(self.phases)

    def get_values(self):
        """
        Returns the wye-connected power, current and admittance
        :return: P0, Q0, ExpP, ExpQ, V0 to conform the exponential model
        """

        if self.number_of_phases == 3:
            if self.conn_type == Connection.Wye:
                return self.P0, self.Q0, self.exp_p, self.exp_p, self.V0
            else:
                return delta_to_wye(self.P0), delta_to_wye(self.Q0), self.exp_p, self.exp_p, self.V0
        else:
            return self.P0, self.Q0, self.exp_p, self.exp_p, self.V0




# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.

from io import StringIO, TextIOWrapper
import os
import chardet
from random import randint, seed
import pandas as pd
import zipfile
from typing import List, Dict

from GridCal.Engine.IO.generic_io_functions import parse_config_df


def save_data_frames_to_zip(dfs: Dict[str, pd.DataFrame], filename_zip="file.zip",
                            text_func=None, progress_func=None):
    """
    Save a list of DataFrames to a zip file without saving to disk the csv files
    :param dfs: dictionary of pandas dataFrames {name: DataFrame}
    :param filename_zip: file name where to save all
    :param text_func: pointer to function that prints the names
    :param progress_func: pointer to function that prints the progress 0~100
    """

    n = len(dfs)

    # open zip file for writing
    with zipfile.ZipFile(filename_zip, 'w', zipfile.ZIP_DEFLATED) as myzip:

        # for each DataFrame and name...
        i = 0
        for name, df in dfs.items():

            # compose the csv file name
            filename = name + ".csv"

            if text_func is not None:
                text_func('Flushing ' + name + ' to ' + filename_zip + '...')

            if progress_func is not None:
                progress_func((i + 1) / n * 100)

            # open a string buffer
            with StringIO() as buffer:

                # save the DataFrame to the buffer
                df.to_csv(buffer, index=False)

                # save the buffer to the zip file
                myzip.writestr(filename, buffer.getvalue())

            i += 1

    print('All DataFrames flushed to zip!')


def get_frames_from_zip(file_name_zip, text_func=None, progress_func=None):
    """
    Open the csv files from a zip file
    :param file_name_zip: name of the zip file
    :param text_func: pointer to function that prints the names
    :param progress_func: pointer to function that prints the progress 0~100
    :return: list of DataFrames
    """

    # open the zip file
    try:
        zip_file_pointer = zipfile.ZipFile(file_name_zip)
    except zipfile.BadZipFile:
        return None

    names = zip_file_pointer.namelist()

    n = len(names)
    data = dict()

    # for each file in the zip file...
    for i, file_name in enumerate(names):

        # split the file name into name and extension
        name, extension = os.path.splitext(file_name)

        if text_func is not None:
            text_func('Unpacking ' + name + ' from ' + file_name_zip)

        if progress_func is not None:
            progress_func((i + 1) / n * 100)

        if extension == '.csv':

            # create a buffer to read the file
            file_pointer = zip_file_pointer.open(file_name)

            if name.lower() == "config":
                try:
                    df = pd.read_csv(file_pointer, index_col=0)
                    data = parse_config_df(df, data)
                except EOFError:
                    return None
                except zipfile.BadZipFile:
                    return None
            else:
                # make pandas read the file
                try:
                    df = pd.read_csv(file_pointer)
                except EOFError:
                    return None
                except zipfile.BadZipFile:
                    return None

            # append the DataFrame to the list
            data[name] = df

    return data


def get_xml_content(file_ptr):
    """
    Reads the content of a file
    :param file_ptr: File pointer (from file or zip file)
    :return: list of text lines
    """
    # xml files always have the encoding declared, find it out
    first_line = file_ptr.readline()

    if b'encoding' in first_line:
        encoding = first_line.split()[2].split(b'=')[1].replace(b'"', b'').replace(b'?>', b'').decode()
    else:
        try:
            detection = chardet.detect(first_line)
            encoding = detection['encoding']
        except:
            encoding = 'utf-8'

    # sequential back to the start
    file_ptr.seek(0)

    # read all the lines
    with TextIOWrapper(file_ptr, encoding=encoding) as fle:
        text_lines = [l for l in fle]

    return text_lines


def get_xml_from_zip(file_name_zip, text_func=None, progress_func=None):
    """
    Get the .xml files from a zip file
    :param file_name_zip: name of the zip file
    :param text_func: pointer to function that prints the names
    :param progress_func: pointer to function that prints the progress 0~100
    :return: list of xml file contents
    """

    # open the zip file
    try:
        zip_file_pointer = zipfile.ZipFile(file_name_zip)
    except zipfile.BadZipFile:
        return None

    names = zip_file_pointer.namelist()

    n = len(names)
    data = dict()

    # for each file in the zip file...
    for i, file_name in enumerate(names):

        # split the file name into name and extension
        name, extension = os.path.splitext(file_name)

        if text_func is not None:
            text_func('Unpacking ' + name + ' from ' + file_name_zip)

        if progress_func is not None:
            progress_func((i + 1) / n * 100)

        if extension == '.xml':
            file_ptr = zip_file_pointer.open(file_name)

            text_lines = get_xml_content(file_ptr)

            data[name] = text_lines

    return data


if __name__ == '__main__':

    # Generate some random values to put in the csv file.
    # seed(42)  # Causes random numbers always be the same for testing.
    # data = [[randint(0, 100) for _ in range(10)] for _ in range(10)]
    # df1 = pd.DataFrame(data)
    #
    # seed(44)  # Causes random numbers always be the same for testing.
    # data = [[randint(0, 100) for _ in range(10)] for _ in range(10)]
    # df2 = pd.DataFrame(data)
    #
    # # save
    # save_data_frames_to_zip({'Data1': df1, 'Data2': df2}, 'some_file.gridcal')
    #
    # # read and print
    # df_list = get_frames_from_zip(file_name_zip='some_file.gridcal')
    #
    # for name, df in df_list.items():
    #     print()
    #     print(name)
    #     print(df)

    fname = r'C:\Users\penversa\Documents\Grids\CGMES\TYNDP_2025\2025NT_ES_model_003.zip'
    data = get_xml_from_zip(fname)

    print()


"""
This helm version is implemented from the thesis of Muthu Subramanian

Implementation by Santiago Peñate Vera 2017
"""

import time
import numpy as np
np.set_printoptions(linewidth=320)
from numpy import zeros, ones, mod, conj, array, r_, linalg, Inf, complex128, c_, r_, angle
from itertools import product
from numpy.linalg import solve
from scipy.sparse.linalg import factorized
from scipy.sparse import issparse, csc_matrix as sparse
import pandas as pd

# Set the complex precision to use
complex_type = complex128


# @jit(cache=True)
def pre_process(n_bus, Yseries, Vset, pq, pv, vd):
    """
    Make the Helm System matrix
    @param n_bus: Number of buses of the circuit
    @param Yseries: Circuit admittance matrix of the series elements
    @param Vset: Vector of voltages of those nodes where the voltage is controlled (AKA Slack and PV buses)
    @param pq: list of PQ node indices
    @param pv: list of PV node indices
    @param vd: list of Slack node indices
    @return: 
    """
    npq = len(pq)
    npv = len(pv)
    nvd = len(vd)
    npqpv = npq + npv

    # now to have efficient arrays of coefficients, we create the array maps
    map_pqpv = zeros(n_bus, dtype=np.int)
    map_w = zeros(n_bus, dtype=np.int)
    map_pqpv[pq] = array(range(npq))
    map_pqpv[pv] = array(range(npv))
    map_w[r_[pq, pv]] = array(range(npqpv))

    # build the expanded system matrix
    Ysys = zeros((2*n_bus, 2*n_bus))

    # Yseries is a CRC sparse matrix, I pass it to coordinates to be able to vectorize the
    m = Yseries.tocoo()
    a = m.row
    b = m.col
    Ysys[2 * a, 2 * b] = m.data.real
    Ysys[2 * a, 2 * b + 1] = -m.data.imag
    Ysys[2 * a + 1, 2 * b] = m.data.imag
    Ysys[2 * a + 1, 2 * b + 1] = m.data.real

    # set pv columns
    Ysys[:, 2 * pv] = zeros((2 * n_bus, npv))
    Ysys[pv * 2 + 1, pv * 2] = ones(npv)

    # set vd elements
    Ysys[vd * 2, :] = zeros((nvd, 2 * n_bus))
    Ysys[vd * 2 + 1, :] = zeros((nvd, 2 * n_bus))
    Ysys[vd * 2, vd * 2] = ones(nvd)
    Ysys[vd * 2 + 1, vd * 2 + 1] = ones(nvd)

    # build the PV matrix
    Ysys_pv = zeros((2 * n_bus, npv))
    for a, b in product(r_[pq, pv], pv):
        kk = map_pqpv[b]
        Ysys_pv[2 * a, kk] = Yseries[a, b].real
        Ysys_pv[2 * a + 1, kk] = Yseries[a, b].imag

    # compute the voltage squared array
    Vset2 = Vset * Vset

    return sparse(Ysys), Ysys_pv, Vset2, map_pqpv, map_w, npq, npv, nvd


def RHS(n, nbus, npq, npv, nvd, Ysh, Ysys_pv, S, Vset, Vset_abs2, C, W, Q, pq, pv, vd, map_pv, map_w):
    """
    Compute the right hand side vector to solve the voltage and reactive power coefficients using the
    factorization of Ysys
    @param n: Order of the coefficients
    @param nbus: Number of buses (not counting the slack buses)
    @param npq: number of PQ nodes
    @param npv: number of PV nodes
    @param nvd: number of slack nodes
    @param Ysh: Vector of the shunt admittance matrix elements
    @param Ysys_pv: System matrix of the PV nodes: Needed to compute the real voltages vector
    @param S: Vector of power injections (nbus elements)
    @param Vset: Vector of set voltages
    @param Vset_abs2: Vector with the voltage set points squared. (nbus elements)
    @param C: Voltage coefficients (Ncoeff x nbus elements)
    @param W: Inverse coefficients structure (Ncoeff x nbus elements)
    @param Q: Reactive power coefficients (Ncoeff x nbus elements)
    @param pq: array of PQ indices
    @param pv: array of PV indices
    @param vd: array of Slack indices
    @param map_pv: map to the indices of the PV dependent structures
    @param map_w: map to the indices of the W structure
    @return: Right hand side vector to solve the coefficients (2 * nbus elements)
    """

    rhs = np.empty(2 * nbus)  # right hand side to solve the system matrix of coefficients
    Vre = ones(npv)  # real voltage vector

    # RHS for the PQ nodes (Vectorized)
    val = RHS_PQ(n, npq, pq, Ysh, C, S, W, map_w)
    rhs[2 * pq] = val.real
    rhs[2 * pq + 1] = val.imag

    # RHS for the VD (Slack) nodes (Vectorized)
    val = RHS_VD(n, nvd, vd, Vset)
    rhs[2 * vd] = val.real
    rhs[2 * vd + 1] = val.imag

    # RHS for the PV nodes (Vectorized)
    val = RHS_PV(n, npv, pv, Ysh, S, C, Q, W, map_pv, map_w)
    rhs[2 * pv] = val.real
    rhs[2 * pv + 1] = val.imag

    # correct RHS with the real part of the voltage in the PV nodes (Vectorized)
    kk = map_pv[pv]
    Vre[kk] = calc_Vre(n, npv, pv, C, Vset_abs2).real
    rhs -= Ysys_pv[:, kk].dot(Vre[kk])

    return rhs, Vre


def delta(n, k):
    return n == k  # is 1 for n==k, 0 otherwise


def RHS_VD(n, nvd, vd, Vset):
    """
    Right hand side calculation for the VD (Slack) nodes.
    @param n: Order of the coefficients
    @param nvd: number of slack nodes
    @param vd: array of Slack indices
    @param Vset: set voltage of the nodes
    @return: Right hand side value for slack nodes
    """

    o1 = ones(nvd, dtype=complex_type)
    if n == 0:
        return o1
    else:
        return (Vset[vd] - o1) * delta(n, 1)


def RHS_PQ(n, npq, pq, Ysh, C, S, W, map_w):
    """
    Right hand side calculation for the PQ nodes.
    @param n: Order of the coefficients
    @param npq: number of PQ nodes
    @param pq: array of PQ indices
    @param Ysh: Vector where every elements the sum of the corresponding row of the reduced admittance matrix)
    @param C: Voltage coefficients (Ncoeff x nbus elements)
    @param S: Vector of power injections (nbus elements)
    @param W: Inverse coefficients structure (Ncoeff x nbus elements)
    @param map_w: map to the indices of the W structure
    @return: Right hand side value for the PQ nodes
    """
    if n == 0:
        return zeros(npq)
    else:
        kw = map_w[pq]
        return conj(S[pq]) * conj(W[n - 1, kw]) - Ysh[pq] * C[n - 1, pq]  # ASU version


def calc_W(n, npqpv, pqpv, kw, C, W):
    """
    Calculation of the inverse coefficients W.
    @param n: Order of the coefficients
    @param npqpv: number of pq and pv nodes
    @param pqpv: array with the PQ and PV node indices
    @param kw: indices that correspond to PQPV in the W structure
    @param C: Structure of voltage coefficients (Ncoeff x nbus elements)
    @param W: Structure of inverse voltage coefficients (Ncoeff x nbus elements)
    @return: Array of inverse voltage coefficients for the order n
    """

    if n == 0:
        res = ones(npqpv, dtype=complex_type)
    else:
        # res = zeros(npqpv, dtype=complex_type)
        # for l in range(n):
        #     res -= W[l, kw] * C[n - l, pqpv]

        l = array(range(n))
        res = -(W[:, kw][l, :] * C[:, pqpv][n - l, :]).sum(axis=0)

    res /= conj(C[0, pqpv])

    return res


def RHS_PV(n, npv, pv, Ysh, S, C, Q, W, map_pv, map_w):
    """
    Right hand side for the PV nodes
    @param n: Order of the coefficients
    @param npv:
    @param pv: array of PV indices
    @param Ysh: Vector where every elements the sum of the corresponding row of the shunt admittance matrix
    @param S: Vector of power injections (nbus elements)
    @param C: Voltage coefficients (Ncoeff x nbus elements)
    @param Q: Reactive power coefficients (Ncoeff x nbus elements)
    @param W: Inverse coefficients structure (Ncoeff x nbus elements)
    @param map_pv: map to the indices of the PV dependent structures
    @param map_w: map to the indices of the W structure
    @return: Right hand side value for the pv nodes
    """

    if n == 0:
        return zeros(npv, dtype=complex_type)  # -1j * Q[0, kk] / conj(C[0, k])

    else:
        kk = map_pv[pv]
        kw = map_w[pv]

        # val = zeros(npv, dtype=complex_type)
        # for l in range(1, n):   # this includes the n-1
        #     val += Q[l, kk] * W[n-l, kw].conjugate()

        if n > 1:
            l = array(range(1, n))
            val = (Q[:, kk][l, :] * W[:, kw][n-l, :].conjugate()).sum(axis=0)
        else:
            val = zeros(npv, dtype=complex_type)

        n1 = n-1
        rhs = S[pv].real * W[n1, kw].conjugate() - (1j * val) - Ysh[pv] * C[n1, pv]

    return rhs


def calc_Vre(n, npv, pv, C, Vset_abs2):
    """
    Compute the real part of the voltage for PV nodes
    @param n: Order of the coefficients
    @param npv: number of PV nodes
    @param pv: array of PV indices
    @param C: Voltage coefficients (Ncoeff x nbus elements)
    @param Vset_abs2: Vector with the voltage set points squared. (nbus elements)
    @return: Real part of the voltage for the PV nodes
    """
    if n == 0:
        return ones(npv, dtype=complex_type)
    elif n == 1:
        return 0.5 * (Vset_abs2[pv] - 1) - 0.5 * calc_R(n, pv, C)
    else:
        return zeros(npv, dtype=complex_type)


def calc_R(n, pv, C):
    """
    Voltage convolution coefficients for the PV nodes
    @param n: Order of the coefficients
    @param pv: array of PV indices
    @param C: Voltage coefficients (Ncoeff x nbus elements)
    @return: Convolution coefficient of order n for the bus k
    """

    # result = zeros(npv, dtype=complex_type)
    # for l in range(n+1):
    #     result += C[l, pv] * C[n - l, pv].conjugate()

    l = array(range(n+1))
    result = (C[:, pv][l, :] * C[:, pv][n - l, :].conjugate()).sum(axis=0)

    return result


def epsilon(Sn, n, E):
    """
    Fast recursive Wynn's epsilon algorithm from:
        NONLINEAR SEQUENCE TRANSFORMATIONS FOR THE ACCELERATION OF CONVERGENCE
        AND THE SUMMATION OF DIVERGENT SERIES

        by Ernst Joachim Weniger
    Args:
        Sn: sum of coefficients
        n: order
        E: Coefficients structure copy that is modified in this algorithm

    Returns:

    """
    Zero = complex_type(0)
    One = complex_type(1)
    Tiny = np.finfo(complex_type).min
    Huge = np.finfo(complex_type).max

    E[n] = Sn

    if n == 0:
        estim = Sn
    else:
        AUX2 = Zero

        for j in range(n, 0, -1):  # range from n to 1 (both included)
            AUX1 = AUX2
            AUX2 = E[j-1]
            DIFF = E[j] - AUX2

            if abs(DIFF) <= Tiny:
                E[j-1] = Huge
            else:
                if DIFF == 0:
                    DIFF = Tiny
                E[j-1] = AUX1 + One / DIFF

        if mod(n, 2) == 0:
            estim = E[0]
        else:
            estim = E[1]

    return estim, E


def epsilonV(nbus, Sn, n, E):
    """
    Fast recursive Wynn's epsilon algorithm from:
        NONLINEAR SEQUENCE TRANSFORMATIONS FOR THE ACCELERATION OF CONVERGENCE
        AND THE SUMMATION OF DIVERGENT SERIES

        by Ernst Joachim Weniger
    Args:
        Sn: sum of coefficients (Vector)
        n: order
        E: Coefficients structure copy that is modified in this algorithm (Matrix)

    Returns:

    """
    Zero = zeros(nbus, dtype=complex_type)
    One = ones(nbus, dtype=complex_type)
    Tiny = np.finfo(complex_type).min
    Huge = np.finfo(complex_type).max

    E[n, :] = Sn

    if n == 0:
        estim = Sn
    else:
        AUX2 = Zero

        for j in range(n, 0, -1):  # range from n to 1 (both included)
            AUX1 = AUX2
            AUX2 = E[j-1, :]
            DIFF = E[j, :] - AUX2

            where_tiny = np.where(DIFF <= Tiny)[0]
            where_zero = np.where(DIFF == Zero)[0]

            if len(where_tiny) > 0:
                E[j-1, where_tiny] = Huge
            else:
                if len(where_zero) > 0:
                    DIFF[where_zero] = Tiny
                E[j-1, :] = AUX1 + One / DIFF

        if mod(n, 2) == 0:
            estim = E[0, :]
        else:
            estim = E[1, :]

    return estim, E


def pade_approximation(n, an, s=1):
    """
    Computes the n/2 pade approximant of the series an at the approximation
    point s

    Arguments:
        an: coefficient matrix, (number of coefficients, number of series)
        n:  order of the series
        s: point of approximation

    Returns:
        pade approximation at s
    """
    nn = int(n / 2)
    if mod(nn, 2) == 0:
        nn -= 1

    L = nn
    M = nn

    an = np.ndarray.flatten(an)
    rhs = an[L + 1:L + M + 1]

    C = zeros((L, M), dtype=complex_type)
    for i in range(L):
        k = i + 1
        C[i, :] = an[L - M + k:L + k]

    try:
        b = solve(C, -rhs)  # bn to b1
    except:
        return 0, zeros(L + 1, dtype=complex_type), zeros(L + 1, dtype=complex_type)

    b = r_[1, b[::-1]]  # b0 = 1

    a = zeros(L + 1, dtype=complex_type)
    a[0] = an[0]
    for i in range(L):
        val = complex_type(0)
        k = i + 1
        for j in range(k + 1):
            val += an[k - j] * b[j]
        a[i + 1] = val

    p = complex_type(0)
    q = complex_type(0)
    for i in range(L + 1):
        p += a[i] * s ** i
        q += b[i] * s ** i

    return p / q, a, b


def interprete_solution(nbus, npv, pv, pqvd, x_sol, Vre, map_pv):
    """
    Assign the solution vector individual values to the correct places
    @param nbus: number of system nodes
    @param npv: number of pv nodes
    @param pv: array of PV node indices
    @param pqvd: array of PQ and Slack node indices
    @param x_sol: solution vector to analyze
    @param Vre: Vector or real part of the voltage for the PV nodes
    @param map_pv: mapping array from normal bus index to PV index
    @return: Voltages coefficients and reactive power coefficients for the PV nodes at the order of x_sol
    """
    C = zeros(nbus, dtype=complex_type)
    Q = zeros(npv)

    # set the PQ and Slack nodes
    C[pqvd] = x_sol[2 * pqvd] + 1j * x_sol[2 * pqvd + 1]

    # Set the PV nodes
    kk = map_pv[pv]
    Q[kk] = x_sol[2 * pv]
    C[pv] = Vre[kk] + 1j * x_sol[2 * pv + 1]

    return C, Q


def helm(Y, Ys, Ysh, max_coefficient_count, S, voltage_set_points, pq, pv, vd, eps=1e-3, use_pade=False):
    """
    Run the holomorphic embedding power flow
    @param Y: Circuit complete admittance matrix
    @param Ys: Circuit series elements admittance matrix
    @param Ysh: Circuit shunt elements admittance matrix
    @param max_coefficient_count: Maximum number of voltage coefficients to evaluate (Must be an odd number)
    @param S: Array of power injections matching the admittance matrix size
    @param voltage_set_points: Array of voltage set points matching the admittance matrix size
    @param pq: list of PQ node indices
    @param pv: list of PV node indices
    @param vd: list of Slack node indices
    @param eps: Tolerance
    @param use_pade: Use the Padè approximation? If False the Epsilon algorithm is used
    @return:
        Vred_last: voltage solution
        converged: Converged? True/False
        normF: Solution error
        Scalc: Power computed from the voltage solution
    """

    start = time.time()

    # number of nodes
    nbus = np.shape(Ys)[0]

    # The routines in this script are meant to handle sparse matrices, hence non-sparse ones are not allowed
    assert(issparse(Ys))

    # Make bus type lists combinations that are going to be used later
    pqvd = r_[pq, vd]
    pqvd.sort()
    pqpv = r_[pq, pv]
    pqpv.sort()

    # prepare the arrays
    Ysys, Ypv, Vset, map_pv, map_w, npq, npv, nvd = pre_process(n_bus=nbus, Yseries=Ys, Vset=voltage_set_points,
                                                                pq=pq, pv=pv, vd=vd)
    npqpv = npq + npv

    # declare the matrix of coefficients that will lead to the voltage computation
    C = zeros((0, nbus), dtype=complex_type)

    # auxiliary array for the epsilon algorithm
    E_v = zeros((0, nbus), dtype=complex_type)
    E_q = zeros((0, npv), dtype=complex_type)

    # Declare the inverse coefficients vector
    # (it is actually a matrix; a vector of coefficients per coefficient order)
    W = zeros((0, npq+npv), dtype=complex_type)

    # Reactive power on the PV nodes
    Q = zeros((0, npv), dtype=complex_type)

    # Squared values of the voltage module for the buses that are not of slack type
    Vset_abs2 = abs(voltage_set_points) ** 2

    # progressive calculation of coefficients
    n = 0
    converged = False
    inside_precision = True
    error = list()
    errors_PV_P = list()
    errors_PV_Q = list()
    errors_PQ_P= list()
    errors_PQ_Q = list()
    voltages = list()
    Sn_v = zeros(nbus, dtype=complex_type)
    Sn_q = zeros(npv, dtype=complex_type)
    voltages_vector = zeros(nbus, dtype=complex_type)
    Vred_last = zeros(nbus, dtype=complex_type)

    # LU factorization of the system matrix (Only needed once)
    solve = factorized(Ysys)

    # set the slack indices voltages
    voltages_vector[vd] = voltage_set_points[vd]

    while n <= max_coefficient_count and not converged and inside_precision:

        # Reserve coefficients memory space
        C = np.vstack((C, np.zeros((1, nbus), dtype=complex_type)))
        E_v = np.vstack((E_v, np.zeros((1, nbus), dtype=complex_type)))
        E_q = np.vstack((E_q, np.zeros((1, npv))))
        W = np.vstack((W, np.zeros((1, npq+npv), dtype=complex_type)))
        Q = np.vstack((Q, np.zeros((1, npv), dtype=complex_type)))

        # get the system independent term to solve the coefficients
        rhs, Vre = RHS(n, nbus, npq, npv, nvd, Ysh, Ypv, S, Vset, Vset_abs2, C, W, Q, pq, pv, vd, map_pv, map_w)

        # Solve the linear system to obtain the new coefficients
        x_sol = solve(rhs)

        # assign the voltages and the reactive power values correctly
        C[n, :], Q[n, :] = interprete_solution(nbus, npv, pv, pqvd, x_sol, Vre, map_pv)

        # copy variables for the epsilon algorithm
        if not use_pade:
            E_v[n, :] = C[n, :]
            E_q[n, :] = Q[n, :]
            Sn_v += C[n, :]
            Sn_q += Q[n, :]

        # Update the inverse voltage coefficients W for the non slack nodes (Vectorized)
        kw = map_w[pqpv]  # actual index in the coefficients structure
        W[n, kw] = calc_W(n, npqpv, pqpv, kw, C, W)

        # calculate the reactive power
        for k in pv:
            kk = map_pv[k]
            if use_pade:
                if mod(n, 2) == 0 and n > 2:
                    q, _, _ = pade_approximation(n, Q)
                    S[k] = S[k].real + 1j * q.real
            else:
                q, E_q[:, kk] = epsilon(Sn_q[kk], n, E_q[:, kk])
                S[k] = S[k].real + 1j * q

        # calculate the voltages
        for k in pqpv:
            if use_pade:
                if mod(n, 2) == 0 and n > 2:
                    v, _, _ = pade_approximation(n, C[:, k])
                    voltages_vector[k] = v
            else:
                # pass
                voltages_vector[k], E_v[:, k] = epsilon(Sn_v[k], n, E_v[:, k])

            if np.isnan(voltages_vector[k]):
                print('Maximum precision reached at ', n)
                voltages_vector = Vred_last
                inside_precision = False
                break

        # voltages_vector, E_v = epsilonV(nbus, Sn_v, n, E_v)

        # copy the voltages array to keep the last solution in case we encounter NaN's due to loss of precision
        Vred_last = voltages_vector.copy()

        # Compose the voltage values from the coefficient series
        voltages.append(voltages_vector.copy())

        # Calculate the error and check the convergence
        Scalc = voltages_vector * conj(Y * voltages_vector)
        # complex power mismatch
        power_mismatch = Scalc - S
        # concatenate error by type
        mismatch = r_[power_mismatch[pv].real, power_mismatch[pq].real, power_mismatch[pq].imag]

        # check for convergence
        normF = linalg.norm(mismatch, Inf)

        if npv > 0:
            a = linalg.norm(mismatch[pv].real, Inf)
        else:
            a = 0
        b = linalg.norm(mismatch[pq].real, Inf)
        c = linalg.norm(mismatch[pq].imag, Inf)
        error.append([a, b, c])

        if normF < eps:
            converged = True
        else:
            converged = False

        n += 1  # increase the coefficients order

    # errors_lst = [array(errors), array(errors_PV_P), array(errors_PV_Q), array(errors_PQ_P), array(errors_PQ_Q)]

    end = time.time()
    elapsed = end - start

    err_df = pd.DataFrame(array(error), columns=['PV_real', 'PQ_real', 'PQ_imag'])
    err_df.plot(logy=True)
    print('Err df:\n', err_df)

    return Vred_last, normF


def bifurcation_point(C, pqpv):
    """
    Computes the bifurcation point by solving the Padè approximation for other values of alpha (loading)
    @param C: voltage coefficients structure
    @param pqpv: array of PQ and PV nodes
    @return: Voltage collapse curves and the matching loading values
    """
    npoints = 100
    order_num, bus_num = np.shape(C)

    V = zeros((npoints, bus_num), dtype=complex_type)
    L = zeros((npoints, bus_num))
    for k in pqpv:

        # get the padè approximation polynomials
        _, p, q = pade_approximation(order_num, C[:, k])

        # compute the asymptotes to compute the voltage from zero to the first asymptote
        asymptotes = np.roots(q[::-1])
        asymptotes = np.sort(abs(asymptotes))
        asymptotes = asymptotes[asymptotes > 2]
        print('Asymptotes', asymptotes)

        # the bifurcation point is the first asymptote (in theory, with infinite numerical precision...)
        bifurc_pt = asymptotes[0]

        # compute a number of loading values to evaluate P(alpha) and Q(alpha)
        alpha = np.linspace(1, bifurc_pt, npoints+1)[:-1]

        # Evaluate the p and q polynomials at the computed loading points
        pval = np.polyval(p[::-1], alpha)
        qval = np.polyval(q[::-1], alpha)

        # Compute the voltage: V(alpha) = P(alpha)/Q(alpha)
        V[:, k] = pval / qval
        L[:, k] = alpha

    return V, L


def res_2_df(V, Sbus, tpe):
    """
    Create dataframe to display the results nicely
    :param V: Voltage complex vector
    :param Sbus: Power complex vector
    :param tpe: Types
    :return: Pandas DataFrame
    """
    vm = abs(V)
    va = np.angle(V)

    d = {1: 'PQ', 2: 'PV', 3: 'VD'}

    tpe_str = array([d[i] for i in tpe], dtype=object)
    data = c_[tpe_str, Sbus.real, Sbus.imag, vm, va]
    cols = ['Type', 'P', 'Q', '|V|', 'angle']
    df = pd.DataFrame(data=data, columns=cols)

    return df


if __name__ == '__main__':
    from GridCal.Engine.calculation_engine import MultiCircuit, PowerFlowOptions, PowerFlow, SolverType
    from matplotlib import pyplot as plt

    grid = MultiCircuit()
    # grid.load_file('lynn5buspq.xlsx')
    # grid.load_file('lynn5buspv.xlsx')
    # grid.load_file('IEEE30.xlsx')
    grid.load_file('/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE 14.xlsx')
    # grid.load_file('/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE39.xlsx')
    # grid.load_file('/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/1354 Pegase.xlsx')

    grid.compile()

    circuit = grid.circuits[0]

    print('\nYbus:\n', circuit.power_flow_input.Ybus.todense())
    print('\nSbus:\n', circuit.power_flow_input.Sbus)
    print('\nIbus:\n', circuit.power_flow_input.Ibus)
    print('\nVbus:\n', circuit.power_flow_input.Vbus)
    print('\ntypes:\n', circuit.power_flow_input.types)
    print('\npq:\n', circuit.power_flow_input.pq)
    print('\npv:\n', circuit.power_flow_input.pv)
    print('\nvd:\n', circuit.power_flow_input.ref)

    start_time = time.time()

    # Y, Ys, Ysh, max_coefficient_count, S, voltage_set_points, pq, pv, vd
    v, err = helm(Y=circuit.power_flow_input.Ybus,
                  Ys=circuit.power_flow_input.Yseries,
                  Ysh=circuit.power_flow_input.Yshunt,
                  max_coefficient_count=30,
                  S=circuit.power_flow_input.Sbus,
                  voltage_set_points=circuit.power_flow_input.Vbus,
                  pq=circuit.power_flow_input.pq,
                  pv=circuit.power_flow_input.pv,
                  vd=circuit.power_flow_input.ref,
                  eps=1e-15)

    print('HEM:')
    print("--- %s seconds ---" % (time.time() - start_time))
    print('Results:\n', res_2_df(v, circuit.power_flow_input.Sbus, circuit.power_flow_input.types))
    print('error: \t', err)

    # check the HELM solution: v against the NR power flow
    print('\nNR')
    options = PowerFlowOptions(SolverType.NR, verbose=False, robust=False, tolerance=1e-9, control_q=False)
    power_flow = PowerFlow(grid, options)

    start_time = time.time()
    power_flow.run()
    print("--- %s seconds ---" % (time.time() - start_time))
    vnr = circuit.power_flow_results.voltage

    print('Results:\n', res_2_df(vnr, circuit.power_flow_input.Sbus, circuit.power_flow_input.types))
    print('error: \t', circuit.power_flow_results.error)

    # check
    print('\ndiff:\t', v - vnr)

    plt.show()



# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.
import time
import multiprocessing
from PySide2.QtCore import QThread, Signal

from GridCal.Engine.basic_structures import Logger
from GridCal.Engine.Core.multi_circuit import MultiCircuit
from GridCal.Engine.Simulations.LinearFactors.analytic_ptdf import *

########################################################################################################################
# Optimal Power flow classes
########################################################################################################################


class LinearAnalysisOptions:

    def __init__(self, distribute_slack=True, correct_values=True):
        """
        Power Transfer Distribution Factors' options
        :param distribute_slack:
        """
        self.distribute_slack = distribute_slack
        self.correct_values = correct_values


class LinearAnalysisDriver(QThread):
    progress_signal = Signal(float)
    progress_text = Signal(str)
    done_signal = Signal()
    name = 'PTDF'

    def __init__(self, grid: MultiCircuit, options: LinearAnalysisOptions):
        """
        Power Transfer Distribution Factors class constructor
        @param grid: MultiCircuit Object
        @param options: OPF options
        """
        QThread.__init__(self)

        # Grid to run
        self.grid = grid

        # Options to use
        self.options = options

        # OPF results
        self.results = LinearAnalysisResults(n_br=0,
                                             n_bus=0,
                                             br_names=[],
                                             bus_names=[],
                                             bus_types=[])

        # set cancel state
        self.__cancel__ = False

        self.all_solved = True

        self.elapsed = 0.0

        self.logger = Logger()

    def run(self):
        """
        Run thread
        """
        start = time.time()
        self.progress_text.emit('Analyzing')
        self.progress_signal.emit(0)

        # Run Analysis
        analysis = LinearAnalysis(grid=self.grid,
                                  distributed_slack=self.options.distribute_slack,
                                  correct_values=self.options.correct_values)

        analysis.run()

        self.results = analysis.results

        end = time.time()
        self.elapsed = end - start
        self.progress_text.emit('Done!')
        self.done_signal.emit()

    def get_steps(self):
        """
        Get variations list of strings
        """
        if self.results is not None:
            return [v for v in self.results.bus_names]
        else:
            return list()

    def cancel(self):
        self.__cancel__ = True


if __name__ == '__main__':

    from GridCal.Engine import FileOpen, SolverType

    fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/Lynn 5 Bus pv.gridcal'
    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE39_1W.gridcal'
    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/grid_2_islands.xlsx'
    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/1354 Pegase.xlsx'

    main_circuit = FileOpen(fname).open()

    options = LinearAnalysisOptions()
    simulation = LinearAnalysisDriver(grid=main_circuit, options=options)
    simulation.run()
    ptdf_df = simulation.results.mdl(result_type=ResultTypes.PTDFBranchesSensitivity)

    print(ptdf_df.get_data_frame())





try:
    import mpi4py
    from GridCal.Engine.Replacements.mpiserve import *
except ImportError:
    pass

from GridCal.Engine.Replacements.poap_controller import *
from GridCal.Engine.Replacements.strategy import *
from GridCal.Engine.Replacements.tcpserve import *



# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.

import json
import numpy as np
from GridCal.Engine.Simulations.result_types import ResultTypes
from GridCal.Engine.Simulations.results_model import ResultsModel


class NMinusKTimeSeriesResults:

    def __init__(self, n, ne, nc, time_array, bus_names, branch_names, bus_types):
        """
        TimeSeriesResults constructor
        @param n: number of buses
        @param m: number of branches
        """

        self.name = 'N-1 time series'

        nt = len(time_array)

        self.bus_types = np.zeros(n, dtype=int)

        self.branch_names = branch_names

        self.bus_names = bus_names

        self.bus_types = bus_types

        self.time_array = time_array

        self.worst_flows = np.zeros((nt, ne))

        self.worst_loading = np.zeros((nt, ne))

        self.overload_count = np.zeros((ne, nc), dtype=int)

        self.relative_frequency = np.zeros((ne, nc))

        self.max_overload = np.zeros((ne, nc))

        self.available_results = [ResultTypes.ContingencyFrequency,
                                  ResultTypes.ContingencyRelativeFrequency,
                                  ResultTypes.MaxOverloads,
                                  ResultTypes.WorstContingencyFlows,
                                  ResultTypes.WorstContingencyLoading]

    def get_steps(self):
        return

    def get_results_dict(self):
        """
        Returns a dictionary with the results sorted in a dictionary
        :return: dictionary of 2D numpy arrays (probably of complex numbers)
        """
        data = {
                'overload_count': self.overload_count.tolist(),
                'relative_frequency': self.relative_frequency.tolist(),
                'max_overload': self.max_overload.tolist(),
                'worst_flows': self.worst_flows.tolist(),
                'worst_loading': self.worst_loading.tolist(),
                }
        return data

    def save(self, fname):
        """
        Export as json
        """
        with open(fname, "w") as output_file:
            json_str = json.dumps(self.get_results_dict())
            output_file.write(json_str)

    def mdl(self, result_type: ResultTypes):
        """
        Plot the results
        :param result_type:
        :return:
        """

        index = self.branch_names

        if result_type == ResultTypes.ContingencyFrequency:
            data = self.overload_count
            y_label = '(#)'
            title = 'Contingency count '
            labels = ['#' + x for x in self.branch_names]

        elif result_type == ResultTypes.ContingencyRelativeFrequency:
            data = self.relative_frequency
            y_label = '(p.u.)'
            title = 'Contingency relative frequency '
            labels = ['#' + x for x in self.branch_names]

        elif result_type == ResultTypes.MaxOverloads:
            data = self.max_overload
            y_label = '(#)'
            title = 'Contingency count '
            labels = ['#' + x for x in self.branch_names]

        elif result_type == ResultTypes.WorstContingencyFlows:
            data = self.worst_flows
            y_label = '(MW)'
            title = 'Worst contingency flows '
            labels = self.branch_names
            index = self.time_array

        elif result_type == ResultTypes.WorstContingencyLoading:
            data = self.worst_loading * 100.0
            y_label = '(%)'
            title = 'Worst contingency loading '
            labels = self.branch_names
            index = self.time_array

        else:
            raise Exception('Result type not understood:' + str(result_type))

        # assemble model
        mdl = ResultsModel(data=data,
                           index=index,
                           columns=labels,
                           title=title,
                           ylabel=y_label)
        return mdl




# This file is part of GridCal.g
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.
print('Loading GridCal...')
import os
import sys
import matplotlib
# matplotlib.use('Qt5Agg')
PACKAGE_PARENT = '..'
SCRIPT_DIR = os.path.dirname(os.path.realpath(os.path.join(os.getcwd(), os.path.expanduser(__file__))))
sys.path.append(os.path.normpath(os.path.join(SCRIPT_DIR, PACKAGE_PARENT)))
from GridCal.__version__ import about_msg
from GridCal.Gui.Main.GridCalMain import run


if __name__ == "__main__":
    print(about_msg)
    run(use_native_dialogues=True)



# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.
import numpy as np
from matplotlib import pyplot as plt
from PySide2.QtCore import QThread, Signal
from pySOT.experimental_design import SymmetricLatinHypercube
from pySOT.strategy import SRBFStrategy
from pySOT.surrogate import GPRegressor
from pySOT.optimization_problems import OptimizationProblem
from poap.controller import ThreadController, BasicWorkerThread

from GridCal.Engine.Core.multi_circuit import MultiCircuit
from GridCal.Engine.Simulations.PowerFlow.power_flow_driver import PowerFlowDriver, PowerFlowOptions
from GridCal.Engine.Simulations.PowerFlow.power_flow_worker import single_island_pf
from GridCal.Engine.Simulations.Stochastic.monte_carlo_results import StochasticPowerFlowResults
from GridCal.Engine.Simulations.Stochastic.monte_carlo_driver import make_monte_carlo_input

########################################################################################################################
# Optimization classes
########################################################################################################################


class VoltageOptimizationProblem(OptimizationProblem):
    """

    :ivar dim: Number of dimensions
    :ivar lb: Lower variable bounds
    :ivar ub: Upper variable bounds
    :ivar int_var: Integer variables
    :ivar cont_var: Continuous variables
    :ivar min: Global minimum value
    :ivar minimum: Global minimizer
    :ivar info: String with problem info
    """
    def __init__(self, circuit: MultiCircuit, options: PowerFlowOptions, max_iter=1000, callback=None):
        self.circuit = circuit

        self.options = options

        self.callback = callback

        # initialize the power flow
        self.power_flow = PowerFlowDriver(self.circuit, self.options)

        n = len(self.circuit.buses)
        m = self.circuit.get_branch_number()

        self.max_eval = max_iter

        # the dimension is the number of nodes
        self.dim = n
        self.min = 0
        self.minimum = np.zeros(self.dim)
        self.lb = -15 * np.ones(self.dim)
        self.ub = 20 * np.ones(self.dim)
        self.int_var = np.array([])
        self.cont_var = np.arange(0, self.dim)
        self.info = str(self.dim) + "Voltage collapse optimization"

        # results
        self.results = StochasticPowerFlowResults(n, m, self.max_eval, name='Voltage optimization')

        # compile circuits
        self.numerical_circuit = self.circuit.compile_snapshot()
        self.numerical_input_islands = self.numerical_circuit.compute(ignore_single_node_islands=options.ignore_single_node_islands)

        self.it = 0

    def eval(self, x):
        """
        Evaluate the Ackley function  at x

        :param x: Data point
        :type x: numpy.array
        :return: Value at x
        :rtype: float
        """
        # For every circuit, run the time series
        for numerical_island in self.numerical_input_islands:
            # sample from the CDF give the vector x of values in [0, 1]
            # c.sample_at(x)
            monte_carlo_input = make_monte_carlo_input(numerical_island)
            mc_time_series = monte_carlo_input.get_at(x)

            Y, I, S = mc_time_series.get_at(t=0)

            #  run the sampled values
            # res = self.power_flow.run_at(0, mc=True)
            res = single_island_pf(circuit, Vbus, Sbus, Ibus, options=self.options, logger=self.logger)

            self.results.S_points[self.it, numerical_island.original_bus_idx] = S
            self.results.V_points[self.it, numerical_island.original_bus_idx] = res.voltage[
                numerical_island.original_bus_idx]
            self.results.Sbr_points[self.it, numerical_island.original_branch_idx] = res.If[
                numerical_island.original_branch_idx]
            self.results.loading_points[self.it, numerical_island.original_branch_idx] = res.loading[
                numerical_island.original_branch_idx]

        self.it += 1
        if self.callback is not None:
            prog = self.it / self.max_eval * 100
            self.callback(prog)

        f = abs(self.results.V_points[self.it - 1, :].sum()) / self.dim
        # print(prog, ' % \t', f)

        return f


class Optimize(QThread):
    progress_signal = Signal(float)
    progress_text = Signal(str)
    done_signal = Signal()

    def __init__(self, circuit: MultiCircuit, options: PowerFlowOptions, max_iter=1000):
        """
        Constructor
        Args:
            circuit: Grid to cascade
            options: Power flow Options
            max_iter: max iterations
        """

        QThread.__init__(self)

        self.circuit = circuit

        self.options = options

        self.max_iter = max_iter

        self.__cancel__ = False

        self.problem = None

        self.solution = None

        self.optimization_values = None

    def run(self):
        """
        Run the optimization
        @return: Nothing
        """

        self.problem = VoltageOptimizationProblem(self.circuit,
                                                  self.options,
                                                  self.max_iter,
                                                  callback=self.progress_signal.emit)

        # # (1) Optimization problem
        # # print(data.info)
        #
        # # (2) Experimental design
        # # Use a symmetric Latin hypercube with 2d + 1 samples
        # exp_des = SymmetricLatinHypercube(dim=self.problem.dim, npts=2 * self.problem.dim + 1)
        #
        # # (3) Surrogate model
        # # Use a cubic RBF interpolant with a linear tail
        # surrogate = RBFInterpolant(kernel=CubicKernel, tail=LinearTail, maxp=self.max_eval)
        #
        # # (4) Adaptive sampling
        # # Use DYCORS with 100d candidate points
        # adapt_samp = CandidateDYCORS(data=self, numcand=100 * self.dim)
        #
        # # Use the serial controller (uses only one thread)
        # controller = SerialController(self.objfunction)
        #
        # # (5) Use the sychronous strategy without non-bound constraints
        # strategy = SyncStrategyNoConstraints(worker_id=0,
        #                                      data=self,
        #                                      maxeval=self.max_eval,
        #                                      nsamples=1,
        #                                      exp_design=exp_des,
        #                                      response_surface=surrogate,
        #                                      sampling_method=adapt_samp)
        #
        # controller.strategy = strategy
        #
        # # Run the optimization strategy
        # result = controller.run()
        #
        # # Print the final result
        # print('Best value found: {0}'.format(result.value))
        # print('Best solution found: {0}'.format(np.array_str(result.params[0], max_line_width=np.inf, precision=5,
        #                                                      suppress_small=True)))

        num_threads = 4

        surrogate_model = GPRegressor(dim=self.problem.dim)
        sampler = SymmetricLatinHypercube(dim=self.problem.dim, num_pts=2 * (self.problem.dim + 1))

        # Create a strategy and a controller
        controller = ThreadController()
        controller.strategy = SRBFStrategy(max_evals=self.max_iter,
                                           opt_prob=self.problem,
                                           exp_design=sampler,
                                           surrogate=surrogate_model,
                                           asynchronous=True,
                                           batch_size=num_threads)

        print("Number of threads: {}".format(num_threads))
        print("Maximum number of evaluations: {}".format(self.max_iter))
        print("Strategy: {}".format(controller.strategy.__class__.__name__))
        print("Experimental design: {}".format(sampler.__class__.__name__))
        print("Surrogate: {}".format(surrogate_model.__class__.__name__))

        # Launch the threads and give them access to the objective function
        for _ in range(num_threads):
            worker = BasicWorkerThread(controller, self.problem.eval)
            controller.launch_worker(worker)

        # Run the optimization strategy
        result = controller.run()

        print('Best value found: {0}'.format(result.value))
        print('Best solution found: {0}\n'.format(np.array_str(result.params[0],
                                                               max_line_width=np.inf,
                                                               precision=4, suppress_small=True)))

        self.solution = result.params[0]

        # Extract function values from the controller
        self.optimization_values = np.array([o.value for o in controller.fevals])

        # send the finnish signal
        self.progress_signal.emit(0.0)
        self.progress_text.emit('Done!')
        self.done_signal.emit()

    def plot(self, ax=None):
        """
        Plot the optimization convergence
        """
        clr = np.array(['#2200CC', '#D9007E', '#FF6600', '#FFCC00', '#ACE600', '#0099CC',
                        '#8900CC', '#FF0000', '#FF9900', '#FFFF00', '#00CC01', '#0055CC'])
        if self.optimization_values is not None:
            max_eval = len(self.optimization_values)

            if ax is None:
                f, ax = plt.subplots()
            # Points
            ax.scatter(np.arange(0, max_eval), self.optimization_values, color=clr[6])
            # Best value found
            ax.plot(np.arange(0, max_eval), np.minimum.accumulate(self.optimization_values), color=clr[1],
                    linewidth=3.0)
            ax.set_xlabel('Evaluations')
            ax.set_ylabel('Function Value')
            ax.set_title('Optimization convergence')

    def cancel(self):
        """
        Cancel the simulation
        """
        self.__cancel__ = True
        self.progress_signal.emit(0.0)
        self.progress_text.emit('Cancelled')
        self.done_signal.emit()



